import pandas as pd
import numpy as np


def index_labels(y):
    # Calculate unique values of y and sort them
    unique_values = get_unique_values(y)
    # Create a dictionary with the unique values as keys and their corresponding index as values
    label_dict = {val: i for i, val in enumerate(unique_values)}
    # Create a new Pandas Series with the indexed values of y
    y_indexed = pd.Series([label_dict[val] for val in y])
    # Return the label dictionary and the indexed Series
    return label_dict, y_indexed


def get_unique_values(y):
    try:
        unique = y.unique()
    except:
        unique = np.unique(np.array(y))
    return sorted(unique)



import pandas as pd

from chester.features_engineering.fe_nlp import get_embeddings
from chester.features_engineering.time_series.ts_feature_extraction import TimeSeriesFeatureExtraction
from chester.run.user_classes import TextFeatureExtraction, TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo


class FeatureHandler:
    def __init__(self, column, col_name,
                 feature_type=None,
                 time_series_handler: TimeSeriesHandler = None,
                 text_feature_extraction: TextFeatureExtraction = None,
                 data_info: DataInfo = None):
        self.column = column
        self.feature_type = feature_type
        self.data_info = data_info
        self.time_series_handler = time_series_handler
        self.col_name = col_name
        self.text_feature_extraction = text_feature_extraction

    def handle_numerical(self):
        self.column.name = "num_" + self.column.name
        return self.column, ["num_" + self.col_name]

    def handle_boolean(self):
        column = self.column.astype(int)
        column.name = "is_" + self.column.name
        return column, ["is_" + self.column.name]

    def handle_categorical(self):
        index_col = pd.Series(self.column).astype('category').cat.codes
        index_col.name = "index_" + self.column.name
        # index_col = index_col.replace(-1, 0)
        return index_col, ["index_" + self.column.name]

    def decide_embedding_size(self):
        self.n_unique_words = self.column.nunique()
        self.text_length = self.column.str.len()
        self.n_rows = self.column.shape[0]
        size = None
        if self.n_unique_words < 100 and self.n_rows < 10000:
            size = 30
        elif self.n_unique_words < 1000 and self.n_rows < 100000:
            size = 90
        elif self.n_unique_words < 5000 and self.n_rows < 1000000:
            size = 210
        elif self.n_unique_words < 10000 and self.n_rows < 10000000:
            size = 240
        else:
            size = 300
        return size

    def handle_text(self):
        embedding_size = self.decide_embedding_size()
        print(f"Feature: {self.col_name} calculating {embedding_size} dim embedding")
        embedding_size_method = int(embedding_size / 3)
        data = pd.DataFrame({self.col_name: self.column})

        if self.text_feature_extraction is not None:
            feat_ext = self.text_feature_extraction
            embedding, _ = get_embeddings(
                training_data=data,
                test_data=None,
                split_data=False,
                text_column=self.col_name,
                corex_dim=feat_ext.corex_dim, corex=feat_ext.corex,
                anchor_words=feat_ext.anchor_words, anchor_strength=feat_ext.anchor_strength,
                bow_dim=feat_ext.bow_dim, bow=feat_ext.bow,
                tfidf_dim=feat_ext.tfidf_dim, tfidf=feat_ext.tfidf,
                ngram_range=feat_ext.ngram_range
            )

        else:
            embedding, _ = get_embeddings(training_data=data, split_data=False, text_column=self.col_name,
                                          corex_dim=embedding_size_method, bow_dim=embedding_size_method,
                                          tfidf_dim=embedding_size_method)

        new_col_name_list = [self.col_name + "_" + col for col in embedding.columns]
        new_col_name_dict = dict(zip(embedding.columns, new_col_name_list))
        embedding.rename(columns=new_col_name_dict, inplace=True)
        return embedding, embedding.columns

    def handle_time_series(self):
        ts_fe = TimeSeriesFeatureExtraction(
            time_series_handler=self.time_series_handler,
            data_info=self.data_info,
            col_name=self.col_name,
            column=self.column
        )
        ts_fe.run()
        print("current columns wow", self.data_info.data.columns)
        self.data_info = ts_fe.data_info

    def run(self):
        if self.feature_type == 'time':
            return self.handle_time_series()
        if self.feature_type == 'numeric':
            return self.handle_numerical()
        elif self.feature_type == 'categorical':
            return self.handle_categorical()
        elif self.feature_type == 'boolean':
            return self.handle_boolean()
        elif self.feature_type == 'text':
            return self.handle_text()
        else:
            if not (self.col_name == self.col_name):
                print(
                    f"No appropriate feature handler found for feature {self.col_name}. This feature will be ignored.")
                return None, None



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as stats

from tamtam.ab_info.ab_class import ABInfo
from tamtam.user_class.user_class import TestInfo
from tamtam.utils.column_utils import Columns


class DeltaPValueCalc:
    def __init__(self, ab_info: ABInfo, test_info: TestInfo):
        self.ab_info = ab_info
        self.test_info = test_info
        self.df = self.ab_info.df
        self.side_col = self.ab_info.test_info.get_side_col()[0]
        self.control = "A"
        self.treatment = "B"
        self.trimmed = self.ab_info.trimmed
        self.trimmed_cols = self.ab_info.trimmed_cols
        self.trimmed_col_by_metric = self.ab_info.trimmed_col_by_metric

    @staticmethod
    def weighted_ttest_1samp(data, popmean, weights):
        # Calculate the weighted mean and standard deviation of the sample
        mean = np.average(data, weights=weights)
        std = np.sqrt(np.average((data - mean) ** 2, weights=weights))

        # Calculate the weighted t-test statistic and p-value
        t_stat = (mean - popmean) / (std / np.sqrt(len(data)))
        p_val = stats.t.sf(np.abs(t_stat), len(data) - 1) * 2

        return t_stat, p_val

    def run_single(self, metric_col):
        weight_col = self.test_info.get_weight_col() or Columns.weight
        is_higher_better = self.test_info.is_higher_better  # saying it should be a test >0 or <=0

        # Subset the data and convert to numpy arrays
        data = self.trimmed[[weight_col, metric_col]].to_numpy()

        # Calculate the weighted mean and standard deviation for the single group
        mean = np.average(data[:, 1], weights=data[:, 0])
        std = np.sqrt(np.average((data[:, 1] - mean) ** 2, weights=data[:, 0]))

        # Calculate the t-test statistic and p-value for the null hypothesis that the mean is 0
        if is_higher_better:
            t_stat, p_val = self.weighted_ttest_1samp(data[:, 1], 0, data[:, 0])
        else:
            t_stat, p_val = self.weighted_ttest_1samp(data[:, 1], 0, data[:, 0])
            t_stat = -t_stat

        # Calculate the effect size, MDE, and CI
        if is_higher_better:
            effect_size = mean
            mde = stats.t.ppf(0.8, len(data) - 1) * std / np.sqrt(len(data))
            ci = stats.t.interval(0.95, len(data) - 1, loc=mean, scale=std / np.sqrt(len(data)))
        else:
            effect_size = -mean
            mde = stats.t.ppf(0.8, len(data) - 1) * std / np.sqrt(len(data))
            ci = stats.t.interval(0.95, len(data) - 1, loc=-mean, scale=std / np.sqrt(len(data)))

        results = {
            'metric': metric_col,
            'weighted_mean': mean,
            'weighted_std': std,
            't_statistic': t_stat,
            'p_value': p_val,
            'effect_size': effect_size,
            'mde': mde,
            'ci_lower': ci[0],
            'ci_upper': ci[1]
        }
        return results

    def run_all_metrics(self):
        single_results = []
        for col in self.trimmed_cols:
            single_results.append(self.run_single(col))
        return single_results

    def plot_single_metric(self, col, delta_results):
        trimmed_cols = self.trimmed_col_by_metric[col]
        test_results = delta_results[delta_results.metric.isin(trimmed_cols)]
        self._plot_util(test_results)

    def plot(self, delta_results):
        for col in self.ab_info.get_metric_cols():
            self.plot_single_metric(col, delta_results)

    @staticmethod
    def _plot_util(df):
        # Extract the unique x-axis values from the DataFrame (corresponding to different trim levels)
        x_values = df['metric'].str.rsplit('_', n=1, expand=True)[1].astype(float)

        # Create a figure with 3 rows and 2 columns of subplots
        fig, axs = plt.subplots(3, 2, figsize=(14, 9))
        plt.title("Significant Calculation")

        # Plot the mean and confidence interval in the top left subplot
        axs[0, 0].errorbar(x_values, df['weighted_mean'],
                           yerr=[df['weighted_mean'] - df['ci_lower'], df['ci_upper'] - df['weighted_mean']],
                           label='Metric diff', marker='o')
        axs[0, 0].plot(x_values, df['weighted_mean'], linestyle='-', label='Weighted mean', color='blue')
        axs[0, 0].plot(x_values, df['ci_upper'], linestyle='--', label='CI Upper', color='red')
        axs[0, 0].plot(x_values, df['ci_lower'], linestyle='--', label='CI Lower', color='green')
        axs[0, 0].set_xlabel('Trim level')
        axs[0, 0].set_ylabel(df['metric'].iloc[0])
        axs[0, 0].legend()

        # Plot the p-value in the top right subplot
        axs[0, 1].plot(x_values, df['p_value'], label='p-value', marker='o')
        axs[0, 1].set_xlabel('Trim level')
        axs[0, 1].set_ylabel('p-value')
        axs[0, 1].legend()

        # Plot the MDE in the bottom left subplot
        axs[1, 0].plot(x_values, df['mde'], label='MDE', marker='o')
        axs[1, 0].set_xlabel('Trim level')
        axs[1, 0].set_ylabel('MDE')
        axs[1, 0].legend()

        # Plot the effect size in the bottom right subplot
        axs[1, 1].plot(x_values, df['effect_size'], label='Effect size', marker='o')
        axs[1, 1].set_xlabel('Trim level')
        axs[1, 1].set_ylabel('Effect size')
        axs[1, 1].legend()

        # Plot the standard deviation in the bottom middle subplot
        axs[2, 0].plot(x_values, df['weighted_std'], label='Std(delta(metric))', marker='o')
        axs[2, 0].set_xlabel('Trim level')
        axs[2, 0].set_ylabel('Std(delta(metric))')
        axs[2, 0].legend()

        # Remove any unused subplots
        axs[2, 1].remove()

        # Adjust the spacing between subplots and display the plot
        fig.tight_layout()
        plt.show()
        plt.close()

    def run(self):
        print("Delta, PValue, effect size:")
        pd.set_option('display.max_columns', None)
        pd.set_option('display.max_rows', None)

        single_results = self.run_all_metrics()  # get diff, pvalue
        print("Test Statistics:")
        print(pd.DataFrame(single_results))

        # plot
        self.plot(delta_results=pd.DataFrame(single_results))



from diamond.model_training.train_single import ImageModelTraining


class ImageModelResults:
    def __init__(self, model, train_loss, val_loss, network_parameters,
                 image_model_training: ImageModelTraining,
                 evaluate_predictions):
        self.model = model
        self.train_loss = train_loss
        self.val_loss = val_loss
        self.network_parameters = network_parameters
        self.image_model_training = image_model_training
        self.evaluate_predictions = evaluate_predictions



import random

META_LEARN_EMOJIS = ["🤖", "📊", "🔬", "🚀", "🔍", "📉", "📈"]
FEATURE_STATS_EMOJIS = ["📈", "📉", "📊", "🔍", "💡", "📚", "🔧"]
FEATURE_ENGINEERING_EMOJIS = ["🔧", "🔬", "🔍", "🧰", "💡", "📚", "📊"]
MODEL_PRE_ANALYSIS_EMOJIS = ["📊", "📈", "🔬", "🧰", "💡", "📚", "🔧"]
MODEL_RUN_EMOJIS = ["🤖", "🔥", "🧠", "🚀", "💻", "💡", "🔬"]
POST_MODEL_ANALYSIS_EMOJIS = ["📊", "📉", "🔬", "📚", "💡", "📈", "📉"]
MODEL_WEAKNESSES_EMOJIS = ["🔬", "📉", "💥", "❌", "🔍", "📈", "💡"]
CLEAN_TEXT_EMOJIS = ["🧼", "🧹", "💧", "🧹", "📚", "💡", "🔍"]


def chapter_title(chapter_name: str, prefix=" Chapter: "):
    stars = '*' * (len(chapter_name + prefix) + 12)
    if chapter_name == 'meta learn':
        emoji = random.choice(META_LEARN_EMOJIS)
    elif chapter_name == 'feature statistics':
        emoji = random.choice(FEATURE_STATS_EMOJIS)
    elif chapter_name == 'feature engineering':
        emoji = random.choice(FEATURE_ENGINEERING_EMOJIS)
    elif chapter_name == 'model pre analysis':
        emoji = random.choice(MODEL_PRE_ANALYSIS_EMOJIS)
    elif chapter_name == 'model training':
        emoji = random.choice(MODEL_RUN_EMOJIS)
    elif chapter_name == 'post model analysis':
        emoji = random.choice(POST_MODEL_ANALYSIS_EMOJIS)
    elif chapter_name == 'model weaknesses':
        emoji = random.choice(MODEL_WEAKNESSES_EMOJIS)
    elif chapter_name == 'model weaknesses':
        emoji = random.choice(MODEL_WEAKNESSES_EMOJIS)
    elif chapter_name == 'text cleaning':
        emoji = random.choice(CLEAN_TEXT_EMOJIS)
    else:
        emoji = '🍵'
    message = f'{stars}\n***{prefix} {chapter_name} {emoji} ***\n{stars}'
    return message



import math
import random
from math import floor

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.patches import Patch
from sklearn.manifold import TSNE
from wordcloud import WordCloud

from chester.zero_break.problem_specification import DataInfo


class CategoricalPreModelAnalysis:
    def __init__(self, data_info: DataInfo):
        self.data_info = data_info
        self.cols = list(set(self.data_info.feature_types_val["categorical"]))
        # drop dups
        if self.data_info.data.columns.duplicated().any():
            drop_dup_data = self.data_info.data.loc[:, ~ self.data_info.data.columns.duplicated()]
            self.data = drop_dup_data
            self.data_info.data = drop_dup_data
        self.n_cols = len(self.cols)
        self.target = self.data_info.data[self.data_info.target]
        self.target_df = self.data_info.data[[self.data_info.target]]
        self.data = self.data_info.data[self.cols]
        # calc
        self.cols_sorted = self.sort_by_pvalue()
        self.cols_sorted_with_pvalue = None

    def tsne(self):
        X = self.data.copy()
        X = X.sample(n=min(5000, len(X)))
        target = self.target[X.index]

        X_tsne_3d = TSNE(n_components=3).fit_transform(pd.get_dummies(X))
        X_tsne_2d = X_tsne_3d[:, :2]

        fig = plt.figure(figsize=(16, 10))
        ax1 = plt
        # ax2 = fig.add_subplot(122, projection='3d')
        if self.data_info.problem_type_val in ["Regression"]:
            ax1.hexbin(X_tsne_2d[:, 0], X_tsne_2d[:, 1], C=target, gridsize=50, cmap='viridis', edgecolors='black')
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=self.target, cmap='viridis')
            ax1.title("Visualizing Categorical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Categorical Features and Target with t-SNE (3D)")
        elif self.data_info.problem_type_val in ["Binary regression", "Binary classification"]:
            target_classes = target.unique()
            color_map = {target_class: color for target_class, color in zip(target_classes, ['red', 'blue'])}
            colors = target.apply(lambda x: color_map[x])
            ax1.scatter(X_tsne_2d[:, 0], X_tsne_2d[:, 1], c=colors)
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=colors)
            legend_handles = [Patch(color=color_map[target_class], label=target_class) for target_class in
                              target_classes]
            ax1.title("Visualizing Categorical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Categorical Features and Target with t-SNE (3D)")
            ax1.legend(handles=legend_handles)
        else:  # Multi-class classification
            target_classes = target.unique()
            color_map = {target_class: color for target_class, color in
                         zip(target_classes, plt.cm.rainbow(np.linspace(0, 1, len(target_classes))))}
            ax1.legend(
                handles=[Patch(color=color_map[target_class], label=target_class) for target_class in target_classes])
            # ax2.legend(
            #     handles=[Patch(color=color_map[target_class], label=target_class) for target_class in target_classes])
        plt.show()
        plt.close()

    def sort_by_pvalue(self):
        from sklearn.cluster import KMeans
        from scipy.stats import chi2_contingency
        import warnings
        warnings.simplefilter("ignore")

        problem_type = self.data_info.problem_type_val

        if problem_type in ["Regression"]:
            num_groups = min(floor(self.data_info.rows / 20), 10)
            kmeans = KMeans(n_clusters=num_groups, n_init=10)
            kmeans.fit(self.target_df)
            target_labels = kmeans.labels_

        feature_pvalue_list = []
        for col in self.cols:
            data_col = self.data[col]
            if problem_type == "Regression":
                contingency_table = pd.crosstab(data_col, columns=target_labels)
                chi2, pvalue, _, _ = chi2_contingency(contingency_table)
            else:
                contingency_table = pd.crosstab(index=data_col, columns=self.target)
                chi2, pvalue, _, _ = chi2_contingency(contingency_table)
            feature_pvalue_list.append((col, pvalue))

        sorted_list = sorted(feature_pvalue_list, key=lambda x: x[1], reverse=False)
        self.cols_sorted_with_pvalue = sorted_list
        return [x[0] for x in sorted_list]

    def analyze_pvalue(self, is_plot=True, top_features=10):
        from chester.util import ReportCollector, REPORT_PATH
        rc = ReportCollector(REPORT_PATH)
        self.sort_by_pvalue()
        rc.save_object(self.cols_sorted_with_pvalue[0:50],
                       text="top 50 with lowest partial pvalue for categorical feat based on chi square:")
        if len(self.cols) == 0:
            return None
        if is_plot:
            self.plot_wordcloud_pvalues(self.cols_sorted_with_pvalue)
            if self.n_cols > 50:
                self.plot_histogram_pvalues(self.cols_sorted_with_pvalue)
        print("Pvalues for top categorical features:")
        print(pd.DataFrame(self.cols_sorted_with_pvalue[0:top_features], columns=["feature", "pvalue"]))

    def partial_plot(self, classification_row_percent=True):
        import warnings
        warnings.simplefilter("ignore")
        top_features = 25
        if self.n_cols <= 25:
            sample_features = self.n_cols
            top_features = self.n_cols
        else:
            sample_features = min(50, int(self.n_cols / 2))

        top_feature_names = random.sample(self.cols_sorted[0:sample_features], top_features)
        feature_index = {feature: index for index, feature in enumerate(self.cols_sorted)}
        top_feature_names.sort(key=lambda x: feature_index[x])
        if self.data_info.problem_type_val in ["Binary regression", "Binary classification"]:
            if top_features == 1:
                fig, ax = plt.subplots(figsize=(14, 10))
                if classification_row_percent:
                    plt.suptitle("Partial Plot to Identify Patterns between Sampled Categorical Features and Target\n"
                                 "Showing % from Feature (row)",
                                 fontsize=14, fontweight='bold')
                else:
                    plt.suptitle("Partial Plot to Identify Patterns between Sampled Categorical Features and Target\n"
                                 "Showing % from Target (column)",
                                 fontsize=14, fontweight='bold')
                if classification_row_percent:
                    crosstab = pd.crosstab(self.data[top_feature_names[0]], self.target, normalize='index') * 100
                    crosstab = crosstab[(crosstab.T != 0).any()]
                    crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                    crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                else:
                    crosstab = pd.crosstab(self.data[top_feature_names[0]], self.target, normalize='columns') * 100
                    crosstab = crosstab[(crosstab.T != 0).any()]
                    crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                    crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                sns.heatmap(crosstab, annot=False, cmap="YlGnBu", fmt='g', ax=ax)
                ax.set_ylabel(None)
                ax.set_xlabel(None)
                ax.set_title(top_feature_names[0], fontsize=12, fontweight='bold')
                plt.show()
                plt.close()

            else:
                max_plots = min(9, top_features)
                dim = min(math.ceil(math.sqrt(max_plots)), 2)
                fig, ax = plt.subplots(dim, dim, figsize=(14, 4 + 4 * dim))
                fig.tight_layout()
                if classification_row_percent:
                    plt.suptitle("Partial Plot to Identify Patterns between Sampled Categorical Features and Target\n"
                                 "Showing % from Feature (row)",
                                 fontsize=14, fontweight='bold')
                else:
                    plt.suptitle("Partial Plot to Identify Patterns between Sampled Categorical Features and Target\n"
                                 "Showing % from Target (column)",
                                 fontsize=14, fontweight='bold')
                for i, col in enumerate(top_feature_names):
                    if i >= dim * dim:
                        break
                    ax_i = ax[i // dim, i % dim]
                    if classification_row_percent:
                        crosstab = pd.crosstab(self.data[col], self.target, normalize='index') * 100
                        crosstab = crosstab[(crosstab.T != 0).any()]
                        crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                        crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                    else:
                        crosstab = pd.crosstab(self.data[col], self.target, normalize='columns') * 100
                        crosstab = crosstab[(crosstab.T != 0).any()]
                        crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                        crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                    sns.heatmap(crosstab, annot=False, cmap="YlGnBu", fmt='g', ax=ax_i)
                    ax_i.set_ylabel(None)
                    ax_i.set_xlabel(None)
                    ax_i.set_title(col, fontsize=12, fontweight='bold')
                plt.tight_layout()
                plt.show()
                plt.close()
        if self.data_info.problem_type_val in ["Regression"]:
            max_plots = 9
            top_n = self.data[:top_features].columns
            dim = max(math.floor(math.sqrt(min(max_plots, len(top_n)))), 2)
            fig, ax = plt.subplots(dim, dim, figsize=(20, 4 + 4 * dim))
            target = self.target
            plt.suptitle(
                "Partial Plot to Identify Patterns between Categorical Sampled Features and Target (grouped by kmeans)",
                fontsize=16,
                fontweight='bold')
            grid_size = math.ceil(math.sqrt(top_features))
            num_features = min(grid_size * grid_size, top_features)
            for i, col in enumerate(top_feature_names[:num_features]):
                if i >= dim * dim:
                    break
                plt.subplot(dim, dim, i + 1)
                column = self.data[col].apply(lambda x: "'" + str(x))
                new_df = pd.concat([column, target], axis=1)
                top_values = list(column.value_counts().nlargest(5).index)
                data_filtered = new_df[column.isin(top_values)]
                sns.boxplot(x=col, y=target, data=data_filtered)
                plt.ylabel("Target")
                plt.xlabel("{} Value".format(col))
                plt.subplots_adjust(hspace=0.5, wspace=0.5)
            plt.show()
            plt.close()
        elif self.data_info.problem_type_val in ["Multiclass classification"]:
            max_plots = 16
            top_n = self.data[:max_plots].columns
            dim = max(math.floor(math.sqrt(len(top_n))), 2)
            fig, ax = plt.subplots(dim, dim, figsize=(19, 4 * dim))
            fig.tight_layout()
            plt.suptitle(
                "Heatmap to Show Correlation between Sampled Categorical Features (top 5 categories) and Target",
                fontsize=15,
                fontweight='bold')
            for i, col in enumerate(top_feature_names):
                if i >= dim * dim:
                    break
                plt.subplot(dim, dim, i + 1)
                crosstab = pd.crosstab(self.data[col], self.target, normalize='index') * 100
                crosstab = crosstab[(crosstab.T != 0).any()]
                crosstab = crosstab.loc[:, (crosstab != 0).any(axis=0)]
                crosstab = crosstab.loc[crosstab.sum(axis=1).sort_values(ascending=False).index[:5]]
                sns.heatmap(crosstab, annot=False, cmap="YlGnBu", fmt='g')
                plt.title(col, fontsize=10, fontweight='bold')
            plt.show()
            plt.close()

    @staticmethod
    def plot_histogram_pvalues(features_pvalues):
        """
        Plot histogram of p-values for features.
        :param features_pvalues: List of tuples (column name, pvalue).
        :return: None.
        """
        pvalues = [pvalue for _, pvalue in features_pvalues]
        fig, ax = plt.subplots()
        plt.figure(figsize=(15, 15))
        ax.hist(pvalues, bins=50, edgecolor='k', color='#2ecc71')
        ax.set_title("Histogram of P-values for Categorical Features", fontsize=16)
        ax.set_xlabel("P-value", fontsize=14)
        ax.set_ylabel("Frequency", fontsize=14)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_linewidth(0.5)
        ax.spines['bottom'].set_linewidth(0.5)
        ax.tick_params(axis='both', which='both', labelsize=12)
        plt.show()
        plt.close()

    @staticmethod
    def plot_wordcloud_pvalues(features_pvalues,
                               title="Categorical Features Pvalues Based on Partial Plot"):
        """
        Plot word cloud of features weighted by their p-value.
        :param features_pvalues: List of tuples (column name, pvalue).
        :param title: Title of the plot.
        :return: None.
        """
        features_pvalues = [(feature, 1 - pvalue) for feature, pvalue in features_pvalues]

        n_features = len(features_pvalues)
        width = min(int(900 * n_features / 4), 900)
        height = min(int(500 * n_features / 4), 500)
        wordcloud = WordCloud(width=width, height=height).generate_from_frequencies(dict(features_pvalues))
        plt.imshow(wordcloud)
        plt.axis("off")
        plt.title(title, fontsize=15)
        plt.show()
        plt.close()

    def run(self, is_plot=True):
        if self.n_cols > 1:
            self.analyze_pvalue(is_plot=is_plot)
            if is_plot:
                if 'classification' or 'binary regression' in self.data_info.problem_type_val.lower():
                    self.partial_plot(classification_row_percent=False)
                    self.partial_plot(classification_row_percent=True)
                else:
                    self.partial_plot()
                self.tsne()
        elif self.n_cols == 1:
            self.analyze_pvalue(is_plot=is_plot)
            if is_plot:
                self.partial_plot()
        plt.show()
        plt.close()


def format_df(df, max_value_width=10, ci_max_value_width=15, ci_col="CI"):
    pd.options.display.max_columns = None

    def trim_value(val):
        if len(str(val)) > max_value_width:
            return str(val)[:max_value_width] + "..."
        return str(val)

    def trim_ci_value(val):
        if len(str(val)) > ci_max_value_width:
            return str(val)[:ci_max_value_width] + "..."
        return str(val)

    df_subset = df.drop(ci_col, axis=1)
    df_subset = df_subset.applymap(trim_value)
    df[df_subset.columns] = df_subset
    df[ci_col] = df[ci_col].apply(trim_ci_value)

    return df



import nltk
import pandas as pd
from nltk.corpus import wordnet
from nltk.stem import PorterStemmer, SnowballStemmer, LancasterStemmer
from nltk.stem import WordNetLemmatizer

from chester.util import ReportCollector, REPORT_PATH


def get_stemmer(name=None):
    if name == "porter":
        return PorterStemmer()
    elif name == "snowball":
        return SnowballStemmer("english")
    elif name == "lancaster":
        return LancasterStemmer()
    else:
        return PorterStemmer()


def lemmatize(text, lemmatizer=WordNetLemmatizer()):
    lemmas = []
    pos_tags = nltk.pos_tag(text)
    for word, pos_tag in pos_tags:
        pos = get_wordnet_pos(pos_tag)
        lemma = lemmatizer.lemmatize(word, pos)
        lemmas.append(lemma)
    return lemmas


def get_wordnet_pos(treebank_tag):
    if treebank_tag.startswith('J'):
        return wordnet.ADJ
    elif treebank_tag.startswith('V'):
        return wordnet.VERB
    elif treebank_tag.startswith('N'):
        return wordnet.NOUN
    elif treebank_tag.startswith('R'):
        return wordnet.ADV
    else:
        return wordnet.NOUN


def tokenize(text):
    tokens = nltk.word_tokenize(text)
    return tokens


# get port stemmer
port_stemmer = PorterStemmer()
# get word net lemmatizer
world_ner_lemmatizer = WordNetLemmatizer()


def stem(words, stemmer=port_stemmer):
    stemmed_words = [stemmer.stem(word) for word in words]
    return stemmed_words


def preprocess_df_text(text_column, preprocessing_options: dict):
    print_preprocessing_options(**preprocessing_options)
    return text_column.apply(lambda x: preprocess_text(x, **preprocessing_options))


def print_preprocessing_options(stemmer=None, lemmatizer=None, stem_flag=False, lemmatize_flag=False,
                                tokenize_flag=True):
    print("Preprocessing step:")
    if stem_flag:
        print("Stemming text using {} stemmer".format(stemmer.__class__.__name__))
    if lemmatize_flag:
        print("Lemmatizing text using {} lemmatizer".format(lemmatizer.__class__.__name__))
    if tokenize_flag:
        print("Tokenizing text")


def preprocess_text(text,
                    stemmer=None,
                    lemmatizer=None,
                    stem_flag=False,
                    lemmatize_flag=False,
                    tokenize_flag=True):
    if stem_flag and lemmatize_flag:
        raise ValueError("Both stemmer and lemmatizer cannot be applied. Please choose one.")

    if tokenize_flag:
        tokens = tokenize(text)
    else:
        tokens = text

    if stem_flag:
        stemmer = stemmer if stemmer is not None else PorterStemmer()
        stemmed_words = stem(tokens, stemmer=stemmer)
    else:
        stemmed_words = tokens

    if lemmatize_flag:
        lemmatized_words = lemmatize(stemmed_words,
                                     lemmatizer=lemmatizer if lemmatizer is not None else world_ner_lemmatizer)
    else:
        lemmatized_words = stemmed_words

    return ' '.join(lemmatized_words)


class TextPreprocessor:
    def __init__(self,
                 df: pd.DataFrame = None,
                 text_column: str = 'text',
                 stemmer=None,
                 lemmatizer=None,
                 stem_flag=True,
                 lemmatize_flag=False,
                 tokenize_flag=True):
        self.df = df
        self.text_column = text_column
        self.stemmer = stemmer
        self.lemmatizer = lemmatizer
        self.stem_flag = stem_flag
        self.lemmatize_flag = lemmatize_flag
        self.tokenize_flag = tokenize_flag

    def generate_report(self):
        report_str = ""
        if self.stem_flag:
            if self.stemmer:
                if self.stemmer == "snowball":
                    report_str += "Using Snowball stemmer, "
                else:
                    report_str += f"Using {self.stemmer} stemmer, "
            else:
                report_str += "Stemming text, "
        if self.lemmatize_flag:
            if self.lemmatizer:
                report_str += f"Using {self.lemmatizer} lemmatizer, "
            else:
                report_str += "Lemmatizing text, "
        if self.tokenize_flag:
            report_str += "Tokenizing text, "
        if report_str:
            report_str = report_str[:-2]
            title_to_print = f"The following preprocessing steps will be applied to the column '{self.text_column}': {report_str}."
            print(title_to_print)
            rc = ReportCollector(REPORT_PATH)
            rc.save_text(title_to_print)
        else:
            print("No preprocessing steps selected.")


def preprocess_text_df(text_preprocessor: TextPreprocessor):
    df = text_preprocessor.df
    text_column = text_preprocessor.text_column

    df[text_column] = df[text_column].apply(lambda x: preprocess_text(
        x,
        stemmer=text_preprocessor.stemmer,
        lemmatizer=text_preprocessor.lemmatizer,
        stem_flag=text_preprocessor.stem_flag,
        lemmatize_flag=text_preprocessor.lemmatize_flag,
        tokenize_flag=text_preprocessor.tokenize_flag,
    ))
    return df



import cv2
import random
import numpy as np
import matplotlib.pyplot as plt

from diamond.user_classes import ImagesData, ImagesAugmentationInfo
from scipy.ndimage import rotate, zoom


class ImageAugmentation:
    def __init__(self, images_data: ImagesData, image_augmentation_info: ImagesAugmentationInfo):
        self.images_data = images_data
        self.image_augmentation_info = image_augmentation_info
        # cals
        self.images, self.labels = self.images_data.images, self.images_data.labels
        self.aug_types, self.aug_prop = self.image_augmentation_info.aug_types, self.image_augmentation_info.aug_prop
        self.is_colored = self.images_data.is_colored
        self.image_shape = self.images_data.image_shape
        self.label_dict = self.images_data.label_dict

    def rotate_image(self):
        # Sample images to rotate
        num_images = len(self.images)
        num_to_rotate = int(num_images * self.aug_prop)
        image_indices = random.choices(range(num_images), k=num_to_rotate)
        # Apply rotation to sampled images
        first_image = self.images[0]
        if self.is_colored:
            rotated_images = np.tile(first_image, reps=(num_to_rotate, 1, 1, 1))
            for i, index in enumerate(image_indices):
                image = self.images[index]
                image = np.transpose(image, (1, 2, 0))
                angle = random.randint(-180, 180)
                rotated = rotate(image, angle, reshape=True)

                resized = np.transpose(
                    cv2.resize(rotated, (image.shape[1], image.shape[0]), interpolation=cv2.INTER_LINEAR),
                    (2, 0, 1))
                rotated_images[i] = resized

            # Concatenate rotated images with original images
            self.images = np.concatenate((self.images, rotated_images[:len(image_indices)]), axis=0)
            self.labels = np.concatenate((self.labels, self.labels[image_indices]), axis=0)
            # Check that the number of images and labels is the same
            if len(self.images) != len(self.labels):
                raise ValueError(f"Number of images and labels must be the same, {len(self.images), len(self.labels)}")
        else:
            rotated_images = np.tile(first_image, reps=(num_to_rotate, 1, 1))
            for i, index in enumerate(image_indices):
                image = self.images[index]
                angle = random.randint(-180, 180)
                rotated = rotate(image, angle, reshape=True)
                resized = cv2.resize(rotated, (image.shape[1], image.shape[0]), interpolation=cv2.INTER_LINEAR)
                rotated_images[i] = resized

            # Concatenate rotated images with original images
            self.images = np.concatenate(
                (self.images, rotated_images[:len(image_indices)].reshape(-1, *self.image_shape)), axis=0)
            self.labels = np.concatenate((self.labels, self.labels[image_indices]), axis=0)
            # Check that the number of images and labels is the same
            if len(self.images) != len(self.labels):
                raise ValueError(f"Number of images and labels must be the same, {len(self.images), len(self.labels)}")

    from scipy.ndimage import zoom

    def zoom_image(self):
        if self.is_colored:
            # Retrieve size and number of channels of the first image in the data
            orig_height, orig_width, *_ = self.image_shape
            # Sample images to zoom
            num_images = len(self.images)
            num_to_zoom = int(num_images * self.aug_prop)
            image_indices = random.choices(range(num_images), k=num_to_zoom)

            # Apply zooming to sampled images
            zoomed_images = np.zeros_like(self.images[:num_to_zoom])
            for i, index in enumerate(image_indices):
                image = self.images[index]
                # Resize the image while maintaining the aspect ratio
                zoomed = zoom(image, (random.uniform(0.2, 1.8), random.uniform(0.2, 1.8), 1))
                # Crop or pad the zoomed image to maintain the original shape
                new_height, new_width, _ = zoomed.shape
                if new_height > orig_height:
                    crop_top = (new_height - orig_height) // 2
                    crop_bottom = new_height - orig_height - crop_top
                    zoomed = zoomed[crop_top:new_height - crop_bottom, :, :]
                else:
                    pad_top = (orig_height - new_height) // 2
                    pad_bottom = orig_height - new_height - pad_top
                    zoomed = np.pad(zoomed, ((pad_top, pad_bottom), (0, 0), (0, 0)), 'constant')
                if new_width > orig_width:
                    crop_left = (new_width - orig_width) // 2
                    crop_right = new_width - orig_width - crop_left
                    zoomed = zoomed[:, crop_left:new_width - crop_right, :]
                else:
                    pad_left = (orig_width - new_width) // 2
                    pad_right = orig_width - new_width - pad_left
                    if self.is_colored:
                        zoomed = np.pad(zoomed, ((0, 0), (pad_left, pad_right), (0, 0)), 'constant')
                    else:
                        zoomed = np.pad(zoomed, ((0, 0), (pad_left, pad_right)), 'constant')
                # Convert the image to the original data type
                zoomed = zoomed.astype(image.dtype)
                zoomed_images[i] = zoomed

            # Concatenate zoomed images with original images
            self.images = np.concatenate((self.images, zoomed_images), axis=0)
            self.labels = np.concatenate((self.labels, self.labels[image_indices]), axis=0)

            # Check that the number of images and labels is the same
            if len(self.images) != len(self.labels):
                raise ValueError(f"Number of images and labels must be the same, {len(self.images), len(self.labels)}")
        else:
            # Retrieve size and number of channels of the first image in the data
            orig_height, orig_width = self.image_shape
            # Sample images to zoom
            num_images = len(self.images)
            num_to_zoom = int(num_images * self.aug_prop)
            image_indices = random.choices(range(num_images), k=num_to_zoom)

            # Apply zooming to sampled images
            zoomed_images = np.zeros_like(self.images[:num_to_zoom])
            for i, index in enumerate(image_indices):
                image = self.images[index]
                # Resize the image while maintaining the aspect ratio
                zoomed = zoom(image, (random.uniform(0.2, 1.8), random.uniform(0.2, 1.8)))
                # Crop or pad the zoomed image to maintain the original shape
                new_height, new_width = zoomed.shape
                if new_height > orig_height:
                    crop_top = (new_height - orig_height) // 2
                    crop_bottom = new_height - orig_height - crop_top
                    zoomed = zoomed[crop_top:new_height - crop_bottom, :]
                else:
                    pad_top = (orig_height - new_height) // 2
                    pad_bottom = orig_height - new_height - pad_top
                    zoomed = np.pad(zoomed, ((pad_top, pad_bottom), (0, 0)), 'constant')
                if new_width > orig_width:
                    crop_left = (new_width - orig_width) // 2
                    crop_right = new_width - orig_width - crop_left
                    zoomed = zoomed[:, crop_left:new_width - crop_right]
                else:
                    pad_left = (orig_width - new_width) // 2
                    pad_right = orig_width - new_width - pad_left
                    zoomed = np.pad(zoomed, ((0, 0), (pad_left, pad_right)), 'constant')
                # Convert the image to the original data type
                zoomed = zoomed.astype(image.dtype)
                zoomed_images[i] = zoomed

            # Concatenate zoomed images with original images
            self.images = np.concatenate((self.images, zoomed_images), axis=0)
            self.labels = np.concatenate((self.labels, self.labels[image_indices]), axis=0)

            # Check that the number of images and labels is the same
            if len(self.images) != len(self.labels):
                raise ValueError(f"Number of images and labels must be the same, {len(self.images), len(self.labels)}")

    def run(self):
        if "rotate" in self.aug_types:
            self.rotate_image()
        if "zoom" in self.aug_types:
            self.zoom_image()

        images_data = ImagesData(images=self.images, labels=self.labels,
                                 validation_prop=self.images_data.validation_prop,
                                 image_shape=self.images_data.image_shape,
                                 label_dict=self.label_dict,
                                 for_model_training=self.images_data.for_model_training)
        self.images_data = images_data
        return images_data



class FeatureEliminationUtils:
    def __init__(self, df):
        self.df = df

    def eliminate_single_values_features(self):
        # Find the features with a single value
        single_value_features = []
        for col in self.df.columns:
            if self.df[col].nunique() == 1:
                single_value_features.append(col)

        # Report the features and remove them from the DataFrame
        if single_value_features:
            print(f"Features with a single value: {', '.join(single_value_features)}")
            self.df = self.df.drop(single_value_features, axis=1)

    def eliminate_high_shared_features(self, p=0.7):
        # Find the features with at least p percent of the same values
        high_correlation_features = []
        for col in self.df.columns:
            if self.df[col].value_counts(normalize=True).iloc[0] >= p:
                high_correlation_features.append(col)

        # Report the features and remove them from the DataFrame
        if high_correlation_features:
            print(f"Features with at least {p * 100}% of the same values: {', '.join(high_correlation_features)}")
            self.df = self.df.drop(high_correlation_features, axis=1)

    def eliminate_duplicates_features(self):
        # Find duplicated features
        duplicates = []
        for i, col1 in enumerate(self.df.columns):
            for col2 in self.df.columns[i + 1:]:
                if self.df[col1].equals(self.df[col2]):
                    duplicates.append(col2)

        # Report the duplicates and remove them from the DataFrame
        if duplicates:
            print(f"Duplicated features: {', '.join(duplicates)}")
            self.df = self.df.drop(duplicates, axis=1)

    def run(self):
        self.eliminate_single_values_features()
        self.eliminate_duplicates_features()
        self.eliminate_high_shared_features()
        return self.df



import torch

from diamond.image_caption.utils import load_images_from_numpy
from diamond.user_classes import ImagesData, ImageDescriptionSpec

import os
from transformers import VisionEncoderDecoderModel, ViTImageProcessor, AutoTokenizer
from transformers import pipeline
import timm
from PIL import ImageDraw


class ImageObjectDetection:
    def __init__(self,
                 images_data: ImagesData,
                 plot=True, plot_sample=10,
                 diamond_collector=None):
        self.images_data = images_data
        self.plot = plot
        self.plot_sample = plot_sample
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.label_dict = self.images_data.label_dict
        self.model = None
        # TODO: later add support for labels (labels VS description?)
        self.diamond_collector = {} if diamond_collector is None else diamond_collector
        self.formatted_images = None
        self.format_images()
        self.load_model()

    def load_model(self):
        self.model = get_image_object_detection_model()

    def format_images(self):
        print("=>Formatting images...")
        self.formatted_images = load_images_from_numpy(self.images_data.raw_images)

    def detect_objects(self):
        bounding_boxes = []
        for image in self.formatted_images:
            # TODO: make sure the right format...
            # test on collab first
            bounding_box = self.model(image)
            bounding_boxes.append(bounding_box)
        return bounding_boxes

    def plot_detected_objets(self):
        if self.plot:
            # sample self.plot_sample images from self.images_to_show
            for i in range(len(self.images_data.images_to_show)):
                image = self.images_data.images_to_show[i]
                bounding_boxes = self.diamond_collector["object detection"][i]
                lables = [pred["label"] for pred in bounding_boxes]
                # Draw bounding boxes and labels on original image
                draw = ImageDraw.Draw(image)
                for prediction in bounding_boxes:
                    box = prediction["box"]
                    label = prediction["label"]
                    draw.rectangle([box["xmin"], box["ymin"], box["xmax"], box["ymax"]], outline="black", width=10)
                    draw.text((box["xmin"], box["ymin"] - 20), label, fill="black")

    def run(self):
        self.diamond_collector["object detection"] = self.detect_objects()
        if self.plot:
            self.plot_detected_objets()


def get_image_object_detection_model():
    model_name = "facebook/detr-resnet-50"
    model_revision = "main"
    model = pipeline("object-detection", model=model_name, revision=model_revision)
    return model



import numpy as np
import os

import numpy as np
import pandas as pd
# exec(open("/Users/amitosi/PycharmProjects/chester/diamond/manual_run.py").read())
from PIL import Image

from diamond.run import run
from diamond.user_classes import ImagesAugmentationInfo, ImageModel, ImageModels

# # define directory path
data_dir = '/Users/amitosi/PycharmProjects/chester/chester/data/blood_type'

# # define image size
img_size = (64, 64)
# img_size = (640, 480)
# img_size = (32, 32)
#
# define empty arrays to store images and labels
images = []
labels = []

# loop over the directories containing the images

# loop over the files in the directory
for filename in os.listdir(data_dir):
    if filename.endswith('.jpg'):
        # label = filename.split("/")[-1].split(".")[0]
        # label = ''.join(char for char in label if not char.isdigit())
        # read the image as a PIL.Image object
        img = Image.open(os.path.join(data_dir, filename)).convert('RGB')
        # print(img.size)
        # resize the image
        img = img.resize(img_size)
        # img = img.resize(img.size)
        # convert the image to a NumPy array
        img_array = np.array(img)
        # append the image array to the list of images
        images.append(img_array)
        # append the label to the list of labels

labels = pd.read_csv(data_dir + "/labels.csv").fillna('UNK')['Category']

# # convert the lists to NumPy arrays
images = np.array(images[0:100])
# print(np.unique(labels))
labels = np.array(labels[0:100])

image_shape = (3, 64, 64)
# image_shape = (3, 480, 640)


image_model_list = [
    ImageModel(network_name="EfficientNetB0",
               batch_size=64 * 64 * 16,
               num_epochs=1,
               optimizer_params={'lr': 0.005},
               dropout=0.7)]
image_models = ImageModels(image_model_list=image_model_list)

diamond_collector = run(images=images,
                        image_shape=image_shape,
                        labels=labels,
                        get_image_description=False,
                        is_augment_data=True,
                        image_augmentation_info=ImagesAugmentationInfo(aug_prop=0.1),
                        is_train_model=True, image_models=image_models,
                        is_post_model_analysis=True,
                        plot=True)



from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo

import pandas as pd
from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo
import random


class TimeFrequencyDecider:
    def __init__(self, column,
                 col_name,
                 time_series_handler: TimeSeriesHandler = None,
                 data_info: DataInfo = None):
        self.date_col_name = col_name  # the name of the date col
        self.column = column
        self.time_series_handler = time_series_handler or TimeSeriesHandler()
        self.data_info = data_info
        self.df = self.data_info.data
        self.id_cols = self.time_series_handler.id_cols or []
        self.df[self.date_col_name] = pd.to_datetime(self.df[self.date_col_name])  # convert to datetime
        self.time_between_events = None
        self.calculate_time_between_events()

    def calculate_time_between_events(self):
        # Get the date column and id columns from self.df
        id_cols = self.id_cols

        if id_cols:
            self.df = self.df.sort_values(by=id_cols + [self.date_col_name], ascending=True)
            groups = self.df.groupby(id_cols)
            time_between_events = groups[self.date_col_name].diff().dt.total_seconds()
        else:
            date_col = self.df[self.date_col_name]
            date_col = date_col.sort_values(ascending=True)
            time_between_events = date_col.diff().dt.total_seconds()

        # Filter out any rows with null values in the new column
        self.time_between_events = time_between_events[~time_between_events.isna()]
        return time_between_events[~time_between_events.isna()]

    def _calculate_raw_frequency(self):
        time_between_events = self.time_between_events
        n = len(time_between_events)
        k = int(0.3 * n)  # Number of values to drop
        sorted_x = sorted(time_between_events)
        remaining_values = sorted_x[:-k]  # Drop top k values
        avg = sum(remaining_values) / len(remaining_values)
        return avg

    @staticmethod
    def translate(t):
        intervals = [
            (1, "second"),
            (15, "seconds"),
            (30, "seconds"),
            (60, "minute"),
            (180, "minutes"),
            (300, "minutes"),
            (900, "minutes"),
            (1800, "minutes"),
            (3600, "hour"),
            (10800, "hours"),
            (21600, "hours"),
            (43200, "hours"),
            (86400, "day"),
            (259200, "days"),
            (604800, "week"),
            (1209600, "weeks"),
            (1814400, "weeks"),
            (2419200, "weeks"),
            (2678400, "month"),
            (7776000, "months"),
            (15552000, "months"),
            (31536000, "year"),
            (63072000, "years"),
            (94608000, "years")
        ]

        # Find the first interval with a value greater than t
        for value, measurement in intervals:
            if t < value:
                t = value
                break

        # Calculate the translated value of t
        # ["second", "seconds","minute", "minutes" ,"hour", "hours","day", "days", "week",
        # "weeks","month", "months", "year", "years"]
        if measurement in ("second", "seconds"):
            translated_t = t
        elif measurement in ("minute", "minutes"):
            translated_t = t / 60
        elif measurement in ("hour", "hours"):
            translated_t = t / 3600
        elif measurement in ("day", "days"):
            translated_t = t / 86400
        elif measurement in ("week", "weeks"):
            translated_t = t / 604800
        elif measurement in ("month", "months"):
            translated_t = t / 2592000
        else:
            translated_t = t / 31536000

        return measurement, translated_t

    def run(self):
        raw_freq = self._calculate_raw_frequency()
        return self.translate(raw_freq)

# dates = ["2022-02-15 " + str(i).zfill(2) + ":00:00" for i in range(24)] * 1

# import random
#
# dates = []
# for _ in range(100):
#     # Choose a random hour, minute, and second
#     hour = random.randint(0, 23)
#     minute = random.randint(0, 59)
#     second = random.randint(0, 59)
#
#     # Construct a string in ISO format with the chosen time
#     date_str = f"2022-02-15 {hour:02d}:{minute:02d}:{second:02d}"
#
#     # Add the string to the list of dates
#     dates.append(date_str)
#
# dates = [date if random.random() > 0.1 else None for date in dates]
# # Create a DataFrame with the dates
# data = pd.DataFrame({
#     "date": pd.to_datetime(dates),
#     "value": 1  # The value is not important and can be flat
# })
# data = data.assign(ID=["A"] * (len(data) // 1) + ["A"][:len(data) % 1])
#
# # Create a DataInfo object with the sample DataFrame
# data_info = DataInfo(data=data, target="value")
#
# # Create a TimeSeriesHandler object with the date column
# decider = TimeFrequencyDecider(
#     column=data.date,
#     col_name="date",
#     time_series_handler=TimeSeriesHandler(id_cols=["ID"]),
#     data_info=data_info
# )
#
# df = decider.calculate_time_between_events()
# # print(df)
# # Print the resulting DataFrame with the time between events
# print(decider.run())



import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from scipy.stats import chi2_contingency

tsne_plot_message = """
The t-SNE plot is a visualization tool that is used to reduce the dimensionality of the data 
and visualize the relationships between the features and the target label in a 2D space. 
The plot shows the transformed data points, where each point represents a sample in the dataset 
and its color represents the target label. The plot can help you understand 
the relationships between the features and the target label in a more intuitive way.
"""

correlation_matrix_message = """
The correlation matrix is a tool used to measure the strength and direction of the linear relationship 
between different features in the dataset. The correlation coefficient ranges from -1 to 1, where a value 
of 1 indicates a perfect positive correlation, meaning that as one feature increases, the other feature 
also increases, a value of -1 indicates a perfect negative correlation, meaning that as one feature 
increases, the other feature decreases, and a value of 0 indicates no correlation between the features. 
It's important to note that a high correlation does not necessarily imply causality, it just indicates 
that the two features are related. In this report, we present the correlation matrix of the features 
in the dataset using a heatmap, where a darker color indicates a stronger correlation. It's worth noting 
that the correlation between the feature and target column (if provided) is also presented. In general, 
a correlation coefficient of 0.7 or higher is considered a strong correlation, a coefficient between 
0.3 and 0.7 is considered a moderate correlation, and a coefficient below 0.3 is considered a weak correlation. 
However, it's important to consider the context of the problem and the domain knowledge when interpreting 
the correlation matrix.
"""


class PreModelAnalysis:
    def __init__(self,
                 df: pd.DataFrame = None,
                 target_column: str = 'target',
                 top_n_features: int = 200,
                 correlation_matrix=True,
                 tsne_plot=True,
                 top_n_pairplot=False,
                 chi_square_test_all_features=True
                 ):
        self.df = df
        self.top_n_features = top_n_features
        self.target_column = target_column
        self.is_model = not (not target_column)
        self.correlation_matrix_bool = correlation_matrix
        self.tsne_plot_bool = tsne_plot
        self.top_n_pairplot_bool = top_n_pairplot
        self.chi_square_test_all_features_bool = chi_square_test_all_features

    def generate_report(self):
        report_str = ""
        if self.correlation_matrix_bool:
            report_str += "Generating Correlation Matrix, "
        if self.tsne_plot_bool:
            report_str += "Generating t-SNE Plot, "
        if self.top_n_pairplot_bool:
            report_str += f"Generating top {self.top_n_features} features Pairplot, "
        if self.chi_square_test_all_features_bool:
            report_str += "Running Chi-Square test on all features, "
        if report_str:
            report_str = report_str[:-2]
            print(f"The following EDA steps will be applied: {report_str}.")
        else:
            print("No EDA steps selected.")

    def get_model_type(self, class_threshold: int = 2):
        is_classification, is_regression = (False, False)
        if not self.is_model:
            return is_classification, is_regression
        target_values = self.df[self.target_column]
        if target_values.dtype == object:
            is_classification = True
        elif target_values.nunique() <= class_threshold:
            is_classification, is_regression = (True, True)
        else:
            is_regression = True
        return is_classification, is_regression

    def correlation_matrix(self):
        print(correlation_matrix_message)
        corr = self.df.corr()
        plt.figure(figsize=(10, 8))
        plt.title("Feature correlation matrix")
        sns.heatmap(corr, annot=True)
        plt.show()
        plt.close()
        return corr

    def tsne_plot(self, n_components=2, perplexity=30.0, n_iter=1000):
        print(tsne_plot_message)
        from sklearn.manifold import TSNE
        X = self.df.drop(columns=self.target_column)
        y = self.df[self.target_column]
        tsne = TSNE(n_components=n_components, perplexity=perplexity, n_iter=n_iter)
        X_tsne = tsne.fit_transform(X)
        plt.scatter(X_tsne[:, 0], X_tsne[:, 1], c=y)
        plt.title("t-SNE Plot of Features and Target Label")
        plt.show()
        plt.close()

    def top_n_pairplot(self, N=4, trimming_left=0.05, trimming_right=0.05):
        if trimming_left > 0:
            print(f"left trimming {100 * trimming_left}% ")
        if trimming_right > 0:
            print(f"right trimming {100 * trimming_right}% ")
        import seaborn as sns
        corr = self.df.corr()
        top_n_features = corr.nlargest(N, self.target_column).index
        top_n_features = top_n_features.drop(self.target_column)
        X = self.df[top_n_features]
        X = X.dropna()
        X = X[(X > X.quantile(trimming_left)) & (X < X.quantile(1 - trimming_right))]
        sns.pairplot(X)
        plt.show()
        plt.close()

    def chi_square_test(self, feature):
        from scipy.stats import chi2_contingency
        X = self.df[[feature, self.target_column]]
        X = X.dropna()
        crosstab = pd.crosstab(X[feature], X[self.target_column])
        chi2, p, dof, expected = chi2_contingency(crosstab)
        return p

    def chi_square_test_all_features(self, k=3):
        pd.options.mode.chained_assignment = None
        from sklearn.cluster import KMeans
        import warnings
        warnings.filterwarnings("ignore", category=UserWarning)

        pvalues = {}
        for feature in self.df.columns:
            if feature != self.target_column:
                X = self.df[[feature]]
                kmeans = KMeans(n_clusters=k, n_init=10)
                kmeans.fit(X)
                X[feature + "_cluster"] = kmeans.labels_
                crosstab = pd.crosstab(X[feature + "_cluster"], self.df[self.target_column])
                chi2, p, dof, expected = chi2_contingency(crosstab)
                pvalues[feature] = p
        return pvalues

    def plot_pvalues(self, threshold=20):
        if not self.is_model:
            pass
        pvalues = self.chi_square_test_all_features()
        if len(pvalues) < threshold:
            print(f"Number of features is less than {threshold}. Not enough data for histogram plot.")
            pass
        import matplotlib.pyplot as plt
        plt.hist(list(pvalues.values()), bins=20)
        plt.xlabel("p-values")
        plt.ylabel("Frequency")
        plt.title("Histogram of p-values")
        plt.show()
        plt.close()

    def run(self):
        import subprocess
        if self.top_n_features:
            self.df = self.df[self.select_top_variance_features(self.top_n_features)]
        if self.correlation_matrix_bool:
            self.correlation_matrix()
        if self.tsne_plot_bool:
            try:
                subprocess.run(["python", "tsne_plot.py"], timeout=120)
            except subprocess.TimeoutExpired:
                print("t-SNE plot did not complete within 2 minutes.")
        if self.top_n_pairplot_bool:
            self.top_n_pairplot()
        if self.chi_square_test_all_features_bool:
            self.plot_pvalues()

    def select_top_variance_features(self, n=200):
        variances = self.df.var()
        top_features = variances.sort_values(ascending=False).head(n).index
        return top_features



import numpy as np
import torch
from PIL import Image
from transformers import VisionEncoderDecoderModel, ViTImageProcessor, AutoTokenizer


def load_from_path(image_path):
    return Image.open(image_path)


def load_images_from_path(images_path):
    return [load_from_path(image_path) for image_path in images_path]


def load_from_numpy(numpy_image):
    # Convert the numpy array to a supported data type
    numpy_image = np.array(numpy_image)
    # Create a PIL Image object from the numpy array
    image = Image.fromarray(numpy_image)
    # try:
    #     image = Image.fromarray(numpy_image, mode='RGBA')
    # except:
    #     image = Image.fromarray(numpy_image, mode='L')

    return image


def load_images_from_numpy(numpy_images):
    return [load_from_numpy(image_np) for image_np in numpy_images]


def predict_caption(images):
    print("loading models... might take a while")
    model = VisionEncoderDecoderModel.from_pretrained("./models")
    feature_extractor = ViTImageProcessor.from_pretrained("./models")
    tokenizer = AutoTokenizer.from_pretrained("./models")
    # print("Done!")
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    max_length = 16
    num_beams = 4
    gen_kwargs = {"max_length": max_length, "num_beams": num_beams}

    new_images = []
    for image in images:
        if image.mode != "RGB":
            image = image.convert(mode="RGB")
        new_images.append(image)
    pixel_values = feature_extractor(images=new_images, return_tensors="pt").pixel_values
    pixel_values = pixel_values.to(device)

    output_ids = model.generate(pixel_values, **gen_kwargs)

    preds = tokenizer.batch_decode(output_ids, skip_special_tokens=True)
    preds = [pred.strip() for pred in preds]
    return preds



from typing import List, Tuple

import numpy as np
from sklearn.linear_model import LogisticRegression
from skopt import gp_minimize

from chester.model_training.data_preparation import CVData, ComplexParameterSet


class HyperparameterOptimization:
    def __init__(self, cv_data: CVData, parameter_set: ComplexParameterSet, metrics: List[Tuple[callable, str]]):
        self.cv_data = cv_data
        self.parameter_set = parameter_set
        self.metrics = metrics

    def optimize(optimization_method='Bayesian', optimization_iterations=20, cv_data=CVData,
                 complex_parameter_set=ComplexParameterSet, target_column='target',
                 metrics_list=[("accuracy", "high")]):
        metric_name, metric_direction = metrics_list[0]
        best_params = None
        best_score = None
        if optimization_method == 'Bayesian':
            # Define the function to optimize
            def logistic_regression_score(params):
                lr = LogisticRegression(C=params[0], penalty=params[1])
                score_list = []
                for train, test in cv_data.splits:
                    X_train, y_train = train_data.iloc[train], train_data[target_column].iloc[train]
                    X_test, y_test = train_data.iloc[test], train_data[target_column].iloc[test]
                    lr.fit(X_train, y_train)
                    y_pred = lr.predict(X_test)
                    score = metric_name(y_test, y_pred)
                    score_list.append(score)
                mean_score = np.mean(score_list)
                return -mean_score if metric_direction == "high" else mean_score

            # Define the parameter space
            param_space = [(0.01, 1), ["l1", "l2"]]

            # Run the optimization
            opt_result = gp_minimize(logistic_regression_score, param_space, n_calls=optimization_iterations)
            best_params = opt_result.x
            best_score = -opt_result.fun if metric_direction == "high" else opt_result.fun
        return best_params, best_score



from catboost import CatBoostClassifier, CatBoostRegressor

from chester.zero_break.problem_specification import DataInfo


class CatboostModel:
    def __init__(self, parameters: list, data_info: DataInfo):
        self.parameters = parameters
        self.model_type = data_info.problem_type_val
        if "regression" in self.model_type.lower():
            # check categorical
            self.model = CatBoostRegressor(cat_features=data_info.feature_types_val["categorical"])
        elif "classification" in self.model_type.lower():
            self.model = CatBoostClassifier(cat_features=data_info.feature_types_val["categorical"])

    # def fit_new

    def fit(self, X, y):
        hyperparams = {param.name: param.value for param in self.parameters}
        self.model.set_params(**hyperparams)
        self.model.fit(X, y)

    def retrain(self, X, y):
        self.model.fit(X, y)

    def transform(self, X):
        return self.model.predict(X)

    def predict(self, X):
        return self.model.predict(X)

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)

    def get_params(self):
        return self.parameters



from chester.run.user_classes import TextFeatureExtraction
from chester.text_stats_analysis.data_quality import TextAnalyzer
from chester.text_stats_analysis.smart_text_analyzer import analyze_text_df
from chester.util import ReportCollector, REPORT_PATH
from chester.zero_break.problem_specification import DataInfo


class TextStats:
    def __init__(self, data_info: DataInfo, text_spec: TextFeatureExtraction = None):
        self.data_info = data_info
        self.text_spec = text_spec
        self.cols = self.data_info.feature_types_val["text"]
        self.data = self.data_info.data

    def run(self):
        rc = ReportCollector(REPORT_PATH)
        rc.save_text("Text Statistics:")
        rc.save_text("Extracting Embedding")
        for col in self.cols:
            title_to_print = f"** Analyzing {col} Column"
            print(title_to_print)
            rc.save_text(title_to_print)
            if self.text_spec is None:
                text_analyzer = TextAnalyzer(df=self.data, text_column=col)
            else:
                corex_topics_num = self.text_spec.corex_dim
                corex_anchor_words = self.text_spec.anchor_words
                corex_anchor_strength = self.text_spec.anchor_strength
                text_analyzer = TextAnalyzer(df=self.data, text_column=col, corex_topics=corex_topics_num,
                                             corex_topics_num=corex_topics_num,
                                             corex_anchor_strength=corex_anchor_strength,
                                             corex_anchor_words=corex_anchor_words)

            analyze_text_df(text_analyzer)



import numpy as np
import pandas as pd
import scipy.sparse as ss
from corextopic import corextopic as ct
from sklearn.feature_extraction.text import CountVectorizer


def get_corex_embedding(training_data, test_data=None, text_column='text', ngram_range=(1, 2),
                        n_topics=50, anchor_words=None, anchor_strength=1.6,
                        max_features=10000):
    # Preprocess data
    vectorizer = CountVectorizer(stop_words='english', max_features=max_features, binary=True, ngram_range=ngram_range)
    doc_word = vectorizer.fit_transform(training_data[text_column])

    doc_word = ss.csr_matrix(doc_word)
    feature_names = list(vectorizer.vocabulary_.keys())
    words = list(np.asarray(feature_names))

    # Train model
    topic_model = ct.Corex(n_hidden=n_topics, words=words, max_iter=200, verbose=False, seed=1)
    if anchor_words is None:
        topic_model.fit(doc_word, words=words)
    else:
        print("Anchored Corex")
        topic_model.fit(doc_word, words=words, anchor_strength=anchor_strength, anchors=anchor_words)

    # Get the topic probabilities for the training data
    topic_probs = topic_model.transform(doc_word, details=True)[0]
    # Normalize the topic probabilities
    topic_probs = topic_probs / np.sum(topic_probs, axis=1, keepdims=True)
    # Create a DataFrame of topic probability features
    topic_prob_df = pd.DataFrame(topic_probs, columns=[f"corex_topic_{i + 1}" for i in range(n_topics)])

    if test_data is not None:
        # Preprocess test data
        test_doc_word = vectorizer.transform(test_data[text_column])
        test_doc_word = ss.csr_matrix(test_doc_word)
        # Get the topic probabilities for the test data
        test_topic_probs = topic_model.transform(test_doc_word, details=True)[0]
        # Normalize the topic probabilities
        test_topic_probs = test_topic_probs / np.sum(test_topic_probs, axis=1, keepdims=True)
        # Create a DataFrame of topic probability features for the test data
        test_topic_prob_df = pd.DataFrame(test_topic_probs, columns=[f"corex_topic_{i + 1}" for i in range(n_topics)])

        return topic_prob_df, test_topic_prob_df
    else:
        return topic_prob_df, None



import os
import random
import sys
from typing import List

path = os.path.abspath("TCAP")
sys.path.append(path)

from sklearn.model_selection import KFold, train_test_split


class CVData:
    def __init__(self, train_data, test_data, target_column, folds=5,
                 split_data=False, split_prop=0.2, split_random_state=42):
        # check if nulls in the target
        null_count = int(train_data[target_column].isnull().sum())
        if null_count > 0:
            null_rows = train_data[train_data[target_column].isnull()].index
            not_null_rows = train_data[~train_data[target_column].isnull()].index
            self.train_data = train_data.loc[not_null_rows]
            self.data_to_predict = train_data.loc[null_rows]
        else:
            self.train_data = train_data
        self.test_data = test_data
        if self.test_data is None:
            if split_data:
                self.train_data, self.test_data = train_test_split(self.train_data, test_size=split_prop,
                                                                   random_state=split_random_state)
        self.target_column = target_column
        self.splits = self.cv_preparation(train_data=self.train_data, test_data=self.test_data, k_fold=folds)

    def format_splits(self):
        formatted_splits = []
        for train_index, test_index in self.splits:
            X_train, y_train = self.train_data.drop(self.target_column, axis=1).iloc[train_index], \
                self.train_data.iloc[train_index][self.target_column]
            X_test, y_test = self.train_data.drop(self.target_column, axis=1).iloc[test_index], \
                self.train_data.iloc[test_index][self.target_column]
            formatted_splits.append((X_train, y_train, X_test, y_test))
        return formatted_splits

    @staticmethod
    def cv_preparation(train_data, test_data=None, k_fold=0):
        if k_fold == 0:
            return None
        elif k_fold > 0:
            kf = KFold(n_splits=k_fold)
            splits = [(train, test) for train, test in kf.split(train_data)]
            return splits
        else:
            raise ValueError("k_fold must be a positive integer")


class Parameter:
    def __init__(self, name, value):
        self.name = name
        self.value = value


class ComplexParameter:

    def __init__(self, name: str, options: List):
        self.name = name
        self.options = options
        self.validate_options()

    def validate_options(self):
        # check if options is a valid type (int, float, tuple, list)
        pass

    def sample(self):
        if isinstance(self.options, int):
            return Parameter(self.name, random.randint(0, self.options))
        elif isinstance(self.options, float):
            return Parameter(self.name, random.uniform(0, self.options))
        elif isinstance(self.options, tuple):
            return Parameter(self.name, random.uniform(self.options[0], self.options[1]))
        elif isinstance(self.options, list):
            return Parameter(self.name, random.choice(self.options))


class ComplexParameterSet:
    def __init__(self, parameters: List[ComplexParameter]):
        self.parameters = parameters

    def sample(self):
        return [parameter.sample() for parameter in self.parameters]


def print_report(parameters: List[Parameter]):
    for param in parameters:
        print(f"{param.name}: {param.value}")



from tamtam.ab_info.ab_class import ABInfo
import seaborn as sns
import matplotlib.pyplot as plt


class MetricCorrelation:
    def __init__(self, ab_info: ABInfo):
        # TODO: fix weights
        self.ab_info = ab_info
        self.df = self.ab_info.df  # pandas
        self.side_col = self.ab_info.test_info.get_side_col()[0]  # column string name
        self.control = "A"
        self.treatment = "B"
        self.metrics = self.ab_info.get_metric_cols()  # m1, m2, ... a list of string

    def calculate_correlation(self):
        metrics = self.metrics  # m1, m2, ... a list of string
        cor_mat = self.df[metrics].corr()
        return cor_mat

    def plot_correlation(self):
        cor_mat = self.calculate_correlation()
        sns.heatmap(cor_mat, annot=True, fmt='.2f', cmap='coolwarm')
        plt.title('Metric Correlation Heatmap')
        plt.show()
        plt.close()

    def partial_plots(self):
        df = self.ab_info.df
        metrics = self.ab_info.get_metric_cols()
        lead_metric = self.ab_info.get_lead_metric()
        other_metrics = metrics.copy()
        other_metrics.remove(lead_metric)

        if len(other_metrics) == 1:
            fig, ax = plt.subplots(figsize=(8, 6))
            sns.regplot(x=lead_metric, y=other_metrics[0], data=df, scatter_kws={'alpha': 0.5}, ax=ax)
            ax.set_xlabel(lead_metric)
            ax.set_ylabel(other_metrics[0])
            ax.set_title(f'Comparing lead metric: {lead_metric} with: {other_metrics[0]}')
        else:
            n_cols = 2
            n_rows = (len(other_metrics) + n_cols - 1) // n_cols

            fig, axes = plt.subplots(n_rows, n_cols, figsize=(n_cols * 6, n_rows * 4))
            axes = axes.ravel()

            for i, metric in enumerate(other_metrics):
                sns.regplot(x=lead_metric, y=metric, data=df, scatter_kws={'alpha': 0.5}, ax=axes[i])
                axes[i].set_xlabel(lead_metric)
                axes[i].set_ylabel(metric)
                axes[i].set_title(f'Comparing lead metric: {lead_metric} with: {metric}')

            plt.tight_layout()
            plt.show()

        plt.close()

    def run(self):
        if len(self.metrics) <= 1:
            return None
        print("==========================> Metrics Correlations ")
        self.plot_correlation()
        self.partial_plots()



import math
import random
from math import floor

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.patches import Patch
from pandas.errors import SettingWithCopyWarning
from scipy.stats import chi2_contingency
from sklearn.cluster import KMeans
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.manifold import TSNE
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from wordcloud import WordCloud

from chester.zero_break.problem_specification import DataInfo


class NumericPreModelAnalysis:
    def __init__(self, data_info: DataInfo):
        self.data_info = data_info
        self.cols = list(set(self.data_info.feature_types_val["numeric"]))
        self.n_cols = len(self.cols)
        self.target = self.data_info.data[self.data_info.target]
        self.target_df = self.data_info.data[[self.data_info.target]]
        self.data = self.data_info.data[self.cols]
        # calc
        self.cols_sorted = self.sort_by_pvalue()
        self.cols_sorted_with_pvalue = None

    def tsne(self):
        if self.n_cols in (1, 2):
            return None
        all_cols = list(self.data.columns)
        ts_cols = list(set([name for name in all_cols if
                            name.startswith("num_ts_mm") or name.startswith("num_ts_freq_") or
                            (name.startswith("num_ts_") and (name.endswith("_sin") or name.endswith("_cos")))]))

        not_ts_cols = [col for col in all_cols if col not in ts_cols]
        if len(ts_cols) > 5:
            numeric_cols = not_ts_cols + self.sample_list(ts_cols)
        else:
            numeric_cols = not_ts_cols
        X = self.data[numeric_cols].copy()
        X = X.sample(n=min(5000, len(X)))
        target = self.target[X.index]

        numerical_transformer = SimpleImputer(strategy='median')
        transformer = ColumnTransformer(
            transformers=[
                ('n', numerical_transformer, numeric_cols)
            ])
        pipeline = Pipeline(steps=[('preprocessor', transformer)])
        X = pipeline.fit_transform(X)
        scaler = StandardScaler()
        X = scaler.fit_transform(X)

        X_tsne_3d = TSNE(n_components=3).fit_transform(X)
        X_tsne_2d = X_tsne_3d[:, :2]
        fig = plt.figure(figsize=(12, 12))
        # ax1 = fig.add_subplot(121)
        ax1 = plt
        # ax2 = fig.gca(122, projection='3d')

        if self.data_info.problem_type_val in ["Regression"]:
            ax1.scatter(X_tsne_2d[:, 0], X_tsne_2d[:, 1], c=target, cmap='viridis')
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=self.target, cmap='viridis')
            ax1.title("Visualizing Numerical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Numerical Features and Target with t-SNE (3D)")
        elif self.data_info.problem_type_val in ["Binary regression", "Binary classification"]:
            target_classes = target.unique()
            color_map = {target_class: color for target_class, color in zip(target_classes, ['red', 'blue'])}
            colors = target.apply(lambda x: color_map[x])
            ax1.scatter(X_tsne_2d[:, 0], X_tsne_2d[:, 1], c=colors)
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=colors)
            legend_handles = [Patch(color=color_map[target_class], label=target_class) for target_class in
                              target_classes]
            ax1.title("Visualizing Numerical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Numerical Features and Target with t-SNE (3D)")
            ax1.legend(handles=legend_handles)
        else:  # Multi-class classification
            target_classes = target.unique()
            color_map = {target_class: color for target_class, color in
                         zip(target_classes, plt.cm.rainbow(np.linspace(0, 1, len(target_classes))))}
            colors = target.apply(lambda x: color_map[x])
            plt.scatter(X_tsne_2d[:, 0], X_tsne_2d[:, 1], c=colors)
            # ax2.scatter(X_tsne_3d[:, 0], X_tsne_3d[:, 1], X_tsne_3d[:, 2], c=colors)
            legend_handles = [Patch(color=color_map[target_class], label=target_class) for target_class in
                              target_classes]
            ax1.title("Visualizing Numerical Features and Target with t-SNE (2D)")
            # ax2.set_title("Visualizing Numerical Features and Target with t-SNE (3D)")
            ax1.legend(handles=legend_handles)
        plt.show()
        plt.close()

    @staticmethod
    def sample_list(l):
        if not l:
            return []

        n = len(l)
        min_sample_size = max(5, int(0.2 * n))
        max_sample_size = min(15, n)

        sample_size = random.randint(min_sample_size, max_sample_size)
        sample = random.sample(l, sample_size)

        return sample

    @staticmethod
    def median_imputation(df):
        import warnings
        warnings.filterwarnings("ignore", category=SettingWithCopyWarning)

        median = df.median()
        df.fillna(median, inplace=True)
        return df

    def any_numeric(self):
        return True if len(self.cols) > 0 else False

    def sort_by_pvalue(self):
        import warnings
        warnings.simplefilter("ignore")

        problem_type = self.data_info.problem_type_val

        if problem_type in ["Regression"]:
            num_groups = min(floor(self.data_info.rows / 20), 10)
            kmeans = KMeans(n_clusters=num_groups, n_init=10)
            kmeans.fit(self.target_df)
            target_labels = kmeans.labels_

        feature_pvalue_list = []
        for col in self.cols:
            data_col = self.data[[col]]
            num_groups = min(floor(self.data_info.rows / 20), 10)
            kmeans = KMeans(n_clusters=num_groups, n_init=10)
            if self.data[col].isna().any():
                data_col = self.median_imputation(data_col)
            kmeans.fit(data_col)
            labels = kmeans.labels_
            if problem_type == "Regression":
                contingency_table = pd.crosstab(index=labels, columns=target_labels)
                chi2, pvalue, _, _ = chi2_contingency(contingency_table)
            else:
                contingency_table = pd.crosstab(index=labels, columns=self.target)
                chi2, pvalue, _, _ = chi2_contingency(contingency_table)
            feature_pvalue_list.append((col, pvalue))

        sorted_list = sorted(feature_pvalue_list, key=lambda x: x[1], reverse=False)
        self.cols_sorted_with_pvalue = sorted_list
        return [x[0] for x in sorted_list]

    def analyze_pvalue(self, is_plot=True, top_features=10):
        from chester.util import ReportCollector, REPORT_PATH
        rc = ReportCollector(REPORT_PATH)
        self.sort_by_pvalue()
        rc.save_object(self.cols_sorted_with_pvalue[0:50],
                       text="top 50 with lowest partial pvalue for numerical feat, based on chi square:")
        self.plot_wordcloud_pvalues(self.cols_sorted_with_pvalue)
        if is_plot:
            if self.n_cols > 50:
                print("Numerical Pvalues plot:")
                self.plot_histogram_pvalues(self.cols_sorted_with_pvalue)
        print("Pvalues for Top Numerical Features for Chi Square test:")
        print(pd.DataFrame(self.cols_sorted_with_pvalue[0:top_features], columns=["feature", "pvalue"]))

    @staticmethod
    def plot_histogram_pvalues(features_pvalues):
        """
        Plot histogram of p-values for features.
        :param features_pvalues: List of tuples (column name, pvalue).
        :return: None.
        """
        pvalues = [pvalue for _, pvalue in features_pvalues]
        fig, ax = plt.subplots()
        plt.figure(figsize=(13, 10))
        ax.hist(pvalues, bins=50, edgecolor='k', color='#2ecc71')
        ax.set_title("Histogram of P-values for Numerical Features", fontsize=18)
        ax.set_xlabel("P-value", fontsize=17)
        ax.set_ylabel("Frequency", fontsize=17)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_linewidth(0.5)
        ax.spines['bottom'].set_linewidth(0.5)
        ax.tick_params(axis='both', which='both', labelsize=12)
        plt.show()
        plt.close()

    @staticmethod
    def plot_wordcloud_pvalues(features_pvalues,
                               title="Numeric Features Pvalues Word Cloud Based on Partial Plot"):
        """
        Plot word cloud of features weighted by their p-value.
        :param features_pvalues: List of tuples (column name, pvalue).
        :param title: Title of the plot.
        :return: None.
        """
        features_pvalues = [(feature, 1 - pvalue) for feature, pvalue in features_pvalues]
        wordcloud = WordCloud(width=900, height=500). \
            generate_from_frequencies(dict(features_pvalues))
        plt.title(title, fontsize=15)
        plt.imshow(wordcloud)
        plt.show()
        plt.close()

    def partial_plot(self, classification_row_percent=True):
        import warnings
        warnings.simplefilter("ignore")
        top_features = min(self.n_cols, 25)
        if self.n_cols <= 25:
            sample_features = self.n_cols
        else:
            sample_features = min(2 * 25, int(self.n_cols / 2))
        A = max(top_features, sample_features)
        B = min(top_features, sample_features)
        top_feature_names = random.sample(self.cols_sorted[0:A], B)
        feature_index = {feature: index for index, feature in enumerate(self.cols_sorted)}
        top_feature_names.sort(key=lambda x: feature_index[x])

        dim = math.ceil(math.sqrt(top_features))
        num_rows = math.ceil(top_features / dim)
        if self.data_info.problem_type_val in ["Binary regression"]:
            try:
                fig, ax = plt.subplots(num_rows, dim)
                plt.figure(figsize=(18, 15))
                plt.suptitle("Partial Plot to Identify Patterns between Sampled Numeric Features and Target",
                             fontsize=16, fontweight='bold')
                for i in range(len(top_feature_names)):
                    col = top_feature_names[i]
                    column = self.data[col]
                    target = self.target
                    ax_i = ax[i // dim, i % dim]
                    sns.regplot(x=column, y=target, logistic=True, n_boot=250, y_jitter=.03, ax=ax_i)
                    ax_i.set_xlabel("")
                    ax_i.set_ylabel("")
                    ax_i.set_title(col, fontweight='bold', transform=ax_i.transAxes, y=0.5)
                plt.show()
                plt.close()
            except:
                pass
        elif self.data_info.problem_type_val in ["Regression"]:
            plt.figure(figsize=(18, 15))
            plt.suptitle("Partial Plot to Identify Patterns between Sampled Numeric Features and Target", fontsize=16,
                         fontweight='bold')
            for i, col in enumerate(top_feature_names[:top_features]):
                plt.subplot(num_rows, dim, i + 1)
                column = self.data[col]
                target = self.target
                plt.scatter(column, target)
                plt.xlabel(col)
                plt.ylabel(self.data_info.target)
            plt.show()
            plt.close()
        elif self.data_info.problem_type_val in ["Multiclass classification", "Binary classification"]:
            dim = min(dim, 3)
            fig, axs = plt.subplots(nrows=dim, ncols=dim, figsize=(18, 15))
            if classification_row_percent:
                plt.suptitle("Partial Plot to Identify Patterns between Sampled Numeric Features and Target\n"
                             "Showing % from cluster (row)",
                             fontsize=14, fontweight='bold')
            else:
                plt.suptitle("Partial Plot to Identify Patterns between Sampled Numeric Features and Target\n"
                             "Showing % from Target (column)",
                             fontsize=14, fontweight='bold')
            for i in range(min(len(top_feature_names), dim * dim)):
                if i < dim * dim:
                    row = i // dim
                    col = i % dim
                    ax = axs[row, col]

                    col = top_feature_names[i]
                    data_col = self.data[[col]]
                    num_groups = min(floor(self.data_info.rows / 20), 10)
                    kmeans = KMeans(n_clusters=num_groups, n_init=10)
                    if self.data[col].isna().any():
                        data_col = self.median_imputation(data_col)
                    kmeans.fit(data_col)
                    labels = kmeans.labels_
                    top_5_target_values = self.target.value_counts().index[:5]
                    target_filtered = self.target[self.target.isin(top_5_target_values)]
                    labels_filtered = labels[self.target.isin(top_5_target_values)]
                    contingency_table = pd.crosstab(index=labels_filtered, columns=target_filtered)
                    if classification_row_percent:
                        contingency_table_pct = contingency_table.div(contingency_table.sum(1), axis=0)
                    else:
                        contingency_table_pct = contingency_table.div(contingency_table.sum(0), axis=1)
                    sns.heatmap(contingency_table_pct, annot=False, cmap='Blues', ax=ax)
                    ax.set_ylabel(col, fontsize=12, fontweight='bold')
                    ax.set_xlabel(None)

            plt.show()
            plt.close()

    def run(self, plot=True):
        if self.n_cols > 1:
            pass
            #     self.analyze_pvalue(is_plot=plot)
            #     if ('classification' in self.data_info.problem_type_val.lower()) and plot:
            #         self.partial_plot(classification_row_percent=True)
            #         self.partial_plot(classification_row_percent=False)
            #     else:
            #         self.partial_plot()
            self.tsne()
        elif self.n_cols == 1:
            self.analyze_pvalue()
            if ('classification' in self.data_info.problem_type_val.lower()) and plot:
                self.partial_plot(classification_row_percent=True)
                self.partial_plot(classification_row_percent=False)
            else:
                if plot:
                    self.partial_plot()
        plt.close()


def format_df(df, max_value_width=10, ci_max_value_width=15, ci_col="CI"):
    pd.options.display.max_columns = None

    def trim_value(val):
        if len(str(val)) > max_value_width:
            return str(val)[:max_value_width] + "..."
        return str(val)

    def trim_ci_value(val):
        if len(str(val)) > ci_max_value_width:
            return str(val)[:ci_max_value_width] + "..."
        return str(val)

    df_subset = df.drop(ci_col, axis=1)
    df_subset = df_subset.applymap(trim_value)
    df[df_subset.columns] = df_subset
    df[ci_col] = df[ci_col].apply(trim_ci_value)

    return df



from diamond.model_training.model_results import ImageModelResults
from diamond.model_training.train_single import ImageModelTraining
from diamond.user_classes import ImagesData, ImageModels


class ImageModelsTraining:
    def __init__(self, images_data: ImagesData,
                 image_models: ImageModels):
        self.images_data = images_data
        self.image_models = image_models
        self.image_model_list = image_models.image_model_list

    def run(self):
        image_model_results = []
        for image_model in self.image_model_list:
            image_model_training = ImageModelTraining(images_data=self.images_data, image_model=image_model)
            model, train_loss, val_loss, evaluate_predictions = image_model_training.run()
            image_model_results.append(ImageModelResults(model=model, train_loss=train_loss, val_loss=val_loss,
                                                         network_parameters=image_model.network_parameters,
                                                         image_model_training=image_model_training,
                                                         evaluate_predictions=evaluate_predictions))
        # sort image_model_results by val loss
        sorted_results = sorted(image_model_results, key=lambda x: x.val_loss)
        return sorted_results



import numpy as np
import pandas as pd

from chester.features_engineering.time_series.ts_utils import min_fix, max_fix, mean_fix, median_fix
from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo


class FrequenciesFeatures:
    # the goal is to calculate moving metrics (like average or max) using last n lag values, for different n and metrics
    def __init__(self, column,
                 col_name,
                 time_series_handler: TimeSeriesHandler = None,
                 data_info: DataInfo = None):
        self.date_col_name = col_name  # the name of the date col
        # self.date_column = column
        self.time_series_handler = time_series_handler or TimeSeriesHandler()
        self.data_info = data_info
        self.df = self.data_info.data.sort_values(self.date_col_name)
        self.id_cols = self.time_series_handler.id_cols or []
        self.lag_values = time_series_handler.lag_values
        self.df[self.date_col_name] = pd.to_datetime(self.df[self.date_col_name])  # convert to datetime
        self.time_between_events = None
        self.calculate_time_between_events()
        # self.target = self.df[self.data_info.target]

    def calculate_time_between_events(self):
        # Get the date column and id columns from self.df
        id_cols = self.id_cols

        if id_cols:
            self.df = self.df.sort_values(by=id_cols + [self.date_col_name], ascending=True)
            groups = self.df.groupby(id_cols)
            time_between_events = groups[self.date_col_name].diff().dt.total_seconds()
        else:
            date_col = self.df[self.date_col_name]
            date_col = date_col.sort_values(ascending=True)
            time_between_events = date_col.diff().dt.total_seconds()
        self.time_between_events = time_between_events
        self.df['time_between_events'] = time_between_events

    def collect_last_values(self):
        target = 'time_between_events'  # change to diff

        if not self.id_cols:
            period_back, calculation_types = self.lag_values
            max_lag = max(period_back)
            target = self.df[target]
            shifted_cols = [target.shift(lag + 1).rename(f'ts_lags_{lag + 1}') for lag in range(max_lag)]
            shifted_df = pd.concat(shifted_cols, axis=1)
            last_values = shifted_df.apply(lambda x: x.dropna().tolist(), axis=1).rename('last_values')
            self.df = pd.concat([self.df, last_values], axis=1)
        else:
            groups = self.id_cols
            period_back, calculation_types = self.lag_values
            max_lag = max(period_back)
            shifted_cols = [self.df.groupby(groups)[target].shift(lag + 1).rename(f'ts_lags_{lag + 1}') for lag in
                            range(max_lag)]
            shifted_df = pd.concat(shifted_cols, axis=1)
            last_values = shifted_df.apply(lambda x: x.dropna().tolist(), axis=1).rename('last_values')
            self.df = pd.concat([self.df, last_values], axis=1)

    def calculate_ts_metrics(self):
        period_back, calculation_types = self.lag_values
        target = self.data_info.target
        date_col_name = self.date_col_name
        suffix = "ts_freq"
        names = []

        for pb in period_back:
            # Take the first pb values from the last_values column
            values = self.df['last_values'].apply(lambda x: x[0:pb])

            for ct in calculation_types:
                if ct == 'min':
                    metric_values = values.apply(lambda x: min_fix(x))
                    name = f'{suffix}_{pb}_{ct}_{date_col_name}'
                    names.append(name)
                    self.df[name] = metric_values
                elif ct == 'max':
                    metric_values = values.apply(lambda x: max_fix(x))
                    name = f'{suffix}_{pb}_{ct}_{date_col_name}'
                    names.append(name)
                    self.df[name] = metric_values
                elif ct == 'mean':
                    metric_values = values.apply(lambda x: mean_fix(x))
                    name = f'{suffix}_{pb}_{ct}_{date_col_name}'
                    names.append(name)
                    self.df[name] = metric_values
                elif ct == 'median':
                    metric_values = values.apply(lambda x: median_fix(x))
                    name = f'{suffix}_{pb}_{ct}_{date_col_name}'
                    names.append(name)
                    self.df[name] = metric_values

        # Remove the last_values column from the DataFrame
        self.df.drop('last_values', axis=1, inplace=True)
        self.df.drop('time_between_events', axis=1, inplace=True)
        return names

    def run(self):
        self.collect_last_values()
        names = self.calculate_ts_metrics()
        return self.df, names

# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/day.csv")
# df.rename(columns={'cnt': 'target'}, inplace=True)
# dat_info = DataInfo(data=df, target='target')
# dat_info.calculate()
# # print(dat_info)
#
# # ts_handler = TimeSeriesHandler()
# ts_handler = TimeSeriesHandler(id_cols=['workingday'])
# ma = FrequenciesFeatures(column=df['dteday'], col_name='dteday', data_info=dat_info, time_series_handler=ts_handler)
# ma.run()
# print(ma.df.columns)



from transformers import VisionEncoderDecoderModel, ViTImageProcessor, AutoTokenizer
import torch
from PIL import Image
import numpy as np
import os


def get_image_paths(directory):
    image_paths = []
    for filename in os.listdir(directory):
        if filename.endswith(".jpg"):
            image_path = os.path.join(directory, filename)
            image_paths.append(image_path)
    return image_paths


images_path = "/Users/amitosi/PycharmProjects/chester/chester/data/weather"

image_path_example = "/Users/amitosi/PycharmProjects/chester/chester/data/weather/sunrise33.jpg"
# Convert the PIL image to a numpy array
image_pil = Image.open(image_path_example)
image_np = np.array(image_pil)
print(image_np.shape)
# exec(open("/Users/amitosi/PycharmProjects/chester/diamond/image_caption/example.py").read())
model = VisionEncoderDecoderModel.from_pretrained("nlpconnect/vit-gpt2-image-captioning")
feature_extractor = ViTImageProcessor.from_pretrained("nlpconnect/vit-gpt2-image-captioning")
tokenizer = AutoTokenizer.from_pretrained("nlpconnect/vit-gpt2-image-captioning")

model.save_pretrained("./models")
feature_extractor.save_pretrained("./models")
tokenizer.save_pretrained("./models")

# # Load the pre-trained model and related components from the saved directory
# print("loading models... might take a while")
# model = VisionEncoderDecoderModel.from_pretrained("./models")
# feature_extractor = ViTImageProcessor.from_pretrained("./models")
# tokenizer = AutoTokenizer.from_pretrained("./models")
# # print("Done!")
# device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
# model.to(device)
# max_length = 16
# num_beams = 4
# gen_kwargs = {"max_length": max_length, "num_beams": num_beams}
#
#
# def predict_step(image_paths):
#     images = []
#     count = 0
#     for image_path in image_paths:
#         count = count + 1
#         print("wow")
#         if count > 100:
#             break
#         try:
#             i_image = Image.fromarray(np.uint8(image_paths))
#         except:
#             i_image = Image.open(image_path)
#         print(i_image)
#         try:
#             if i_image.mode != "RGB":
#                 i_image = i_image.convert(mode="RGB")
#         except:
#             pass
#         images.append(i_image)
#
#     pixel_values = feature_extractor(images=images, return_tensors="pt").pixel_values
#     pixel_values = pixel_values.to(device)
#
#     output_ids = model.generate(pixel_values, **gen_kwargs)
#
#     preds = tokenizer.batch_decode(output_ids, skip_special_tokens=True)
#     preds = [pred.strip() for pred in preds]
#     return preds
#
#
# # predictions = predict_step(get_image_paths(images_path))
# predictions = predict_step([image_np])



import pickle

import numpy as np

from diamond.run import run
from diamond.user_classes import ImagesAugmentationInfo, ImageModel, ImageModels

# exec(open("/Users/amitosi/PycharmProjects/chester/diamond/manual_run.py").read())

# define directory path
with open("/Users/amitosi/PycharmProjects/chester/chester/data/data_batch_1", 'rb') as fo:
    data = pickle.load(fo, encoding='bytes')
# subdir: not pizza, .jpg
# subdir: pizza, .jpg

# convert the images and labels arrays to numpy arrays
# Extract the images and labels
images = data[b'data'][0:20]
labels = np.array(data[b'labels'][0:20])

# images = np.array(images[0:50])
# labels = np.array(labels[0:50])


image_shape = (3, 64, 64)

image_model_list = [
    ImageModel(network_name="EfficientNetB0",
               batch_size=64 * 64 * 16,
               num_epochs=1,
               optimizer_params={'lr': 0.005},
               dropout=0.7)]
image_models = ImageModels(image_model_list=image_model_list)

diamond_collector = run(images=images,
                        image_shape=image_shape,
                        labels=labels,
                        get_image_description=False,
                        get_object_detection=True,
                        is_augment_data=False,
                        image_augmentation_info=ImagesAugmentationInfo(aug_prop=0.1),
                        is_train_model=False, image_models=image_models,
                        is_post_model_analysis=False,
                        plot=False)



import numpy as np
import torch
import torch.hub as hub
import torch.nn as nn
from IPython import get_ipython
from torch import optim
from torchvision import models
from tqdm import tqdm

from diamond.user_classes import ImagesData, ImageModel


class ImageModelTraining:
    def __init__(self, images_data: ImagesData,
                 image_model: ImageModel):
        self.images_data = images_data
        self.image_model = image_model
        self.num_classes = len(np.unique(images_data.labels))
        self.num_epochs = image_model.num_epochs
        self.train_loader, self.val_loader = images_data.create_data_loaders(
            batch_size=self.image_model.batch_size)
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        print(f"Training using {self.device}")
        if 'google.colab' in str(get_ipython()) and self.device == "cpu":
            print(
                f"If you are using Google Colab, you may want to change the runtime to use a GPU or TPU by "
                f"selecting 'Runtime' -> 'Change runtime type'.")
        self.evaluate_predictions = None
        self.is_colored = self.images_data.is_colored

    def get_model(self):
        supported_models = ["EfficientNetB0", "EfficientNetB4", "EfficientNetB7", "ResNet50", "ResNet101",
                            "DenseNet121", "DenseNet161", "DenseNet201"]
        if self.image_model.network_name == "EfficientNetB0":
            model = hub.load('rwightman/pytorch-image-models', 'efficientnet_b0', pretrained=True)
        elif self.image_model.network_name == "EfficientNetB4":
            model = hub.load('rwightman/pytorch-image-models', 'efficientnet_b4', pretrained=True)
        elif self.image_model.network_name == "EfficientNetB7":
            model = hub.load('rwightman/pytorch-image-models', 'efficientnet_b7', pretrained=True)
        elif self.image_model.network_name == "ResNet50":
            model = models.resnet50(pretrained=True)
        elif self.image_model.network_name == "ResNet101":
            model = models.resnet101(pretrained=True)
        elif self.image_model.network_name == "DenseNet121":
            model = models.densenet121(pretrained=True)
        elif self.image_model.network_name == "DenseNet161":
            model = models.densenet161(pretrained=True)
        elif self.image_model.network_name == "DenseNet169":
            model = models.densenet169(pretrained=True)
        elif self.image_model.network_name == "DenseNet201":
            model = models.densenet201(pretrained=True)
        else:
            raise ValueError(
                f"Unsupported network name: {self.image_model.network_name}, please choose one of: {supported_models}")
        return model

    def load_model(self):
        model = self.get_model().to(self.device)

        # remove the specified number of layers from the top
        if self.image_model.remove_last_layers_num > 0:
            print(f"Removing last {self.image_model.remove_last_layers_num} layers for {self.image_model.network_name}")
            if "resnet" in self.image_model.network_name.lower():
                num_features = model.fc.in_features
                model.fc = nn.Sequential(
                    nn.Dropout(p=self.image_model.dropout),
                    nn.Linear(num_features, self.num_classes)
                )
            elif "inception" in self.image_model.network_name.lower():
                num_features = model.fc.in_features
                model.fc = nn.Identity()
                model.fc_new = nn.Linear(num_features, self.num_classes)
            else:
                for i in range(self.image_model.remove_last_layers_num):
                    try:
                        num_features = model.classifier.in_features
                    except:
                        num_features = model.classifier[0].in_features

                    model.classifier = nn.Sequential(
                        *list(model.classifier.children())[:-1],
                        nn.Dropout(p=self.image_model.dropout),
                        nn.Linear(num_features, self.num_classes)
                    )

        # make the remaining layers non-trainable
        for param in model.parameters():
            param.requires_grad = True

        return model

    def train_model(self, model, criterion, optimizer):
        train_loss, evaluate_accuracy = 0, 0
        for epoch in tqdm(range(self.num_epochs)):
            train_loss = self.train_model_epoch(model, self.train_loader, criterion, optimizer)
            evaluate_accuracy = self.evaluate_model(model=model, data_loader=self.val_loader, device=self.device)
            tqdm.write(
                f"Epoch {epoch + 1}/{self.num_epochs} - Train Loss: {train_loss:.4f} - Evaluate Accuracy: {evaluate_accuracy:.4f}")
        return train_loss, evaluate_accuracy

    def train_model_epoch(self, model, dataloader, criterion, optimizer):
        running_loss = 0.0
        for i, (inputs_raw, labels) in enumerate(dataloader, 0):
            optimizer.zero_grad()

            if self.is_colored:
                inputs, labels = inputs_raw.to(self.device).float(), labels.to(self.device)
            else:
                # Add a new dimension to the input tensors to represent the grayscale channel
                inputs_gray = inputs_raw.unsqueeze(1)
                inputs = torch.cat([inputs_gray, inputs_gray, inputs_gray], dim=1)
                inputs, labels = inputs.to(self.device).float(), labels.to(self.device)
            try:
                model = model.to(self.device)
                outputs = model(inputs).to(self.device)
            except Exception as e:
                print(f"Error while running model: {e}")
                print("Input size:", inputs.size())
            loss = criterion(outputs, labels)
            loss.backward()

            optimizer.step()

            running_loss += loss.item()

        epoch_loss = running_loss / len(dataloader)

        return epoch_loss

    def evaluate_model(self, model, data_loader, device):
        correct = 0
        total = 0
        with torch.no_grad():
            for data in data_loader:
                images, labels = data

                if self.is_colored:
                    inputs, labels = images.to(device).float(), labels.to(device)
                else:
                    # Add a new dimension to the input tensors to represent the grayscale channel
                    inputs_gray = images.unsqueeze(1)
                    inputs = torch.cat([inputs_gray, inputs_gray, inputs_gray], dim=1)
                    inputs, labels = inputs.to(self.device).float(), labels.to(self.device)

                outputs = model(inputs).to(device)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        accuracy = correct / total
        return accuracy

    def get_eval_prediction(self, model, data_loader, device):
        # Get the predictions for the eval set
        eval_preds = []
        with torch.no_grad():
            for data in data_loader:
                images, labels = data
                images, labels = images.to(device).float(), labels.to(device)

                if self.is_colored:
                    inputs = images.to(device).float()
                else:
                    # Add a new dimension to the input tensors to represent the grayscale channel
                    inputs_gray = images.unsqueeze(1)
                    inputs = torch.cat([inputs_gray, inputs_gray, inputs_gray], dim=1)
                    inputs = inputs.to(self.device).float()

                outputs = model(inputs).to(device)
                for i in range(outputs.size(0)):  # iterate over batch size
                    output = outputs[i]
                    _, predicted = torch.max(output, 0)
                    eval_preds.append(predicted)
        eval_preds = torch.stack(eval_preds)
        return eval_preds

    def run(self):
        model = self.load_model()
        criterion = nn.CrossEntropyLoss()  # future: more loss functions, for bb, regression
        optimizer_params = self.image_model.optimizer_params
        print("\nTraining Specifications: ", self.image_model.network_parameters)
        optimizer = optim.Adam(model.parameters(), **optimizer_params)  # get rid of lr
        train_loss, val_loss = self.train_model(model, criterion, optimizer)  # get rid of epochs
        self.evaluate_predictions = self.get_eval_prediction(model, self.val_loader, device=self.device)
        return model, train_loss, val_loss, self.evaluate_predictions



from itertools import chain

from matplotlib import pyplot as plt

from chester.cleaning.cleaner_handler import CleanerHandler
from chester.feature_stats.categorical_stats import CategoricalStats
from chester.feature_stats.text_stats import TextStats
from chester.feature_stats.time_series_stats import TimeSeriesFeatureStatistics
from chester.features_engineering.features_handler import FeaturesHandler
from chester.features_engineering.time_series.ts_features_extraction import TimeSeriesFeaturesExtraction
from chester.model_monitor.error_prediction import ModelWeaknesses
from chester.model_monitor.model_boostrap import ModelBootstrap
from chester.model_training.models.chester_models.best_model import BestModel
from chester.post_model_analysis.post_model_analysis_class import PostModelAnalysis
from chester.pre_model_analysis.categorical import CategoricalPreModelAnalysis
from chester.pre_model_analysis.numerics import NumericPreModelAnalysis
from chester.pre_model_analysis.target import TargetPreModelAnalysis
from chester.pre_model_analysis.time_series import TimeSeriesPreModelAnalysis
from chester.preprocessing.preprocessor_handler import PreprocessHandler
from chester.run.chapter_titles import chapter_title
from chester.run.user_classes import Data, TextHandler, FeatureStats, ModelRun, TextFeatureExtraction, FeatureTypes, \
    TimeSeriesHandler
from chester.util import REPORT_PATH, ReportCollector
from chester.zero_break.problem_specification import DataInfo

import logging

logging.basicConfig(level=logging.WARNING)

from chester.model_training.data_preparation import CVData
import pandas as pd
from sklearn.preprocessing import LabelEncoder


def run_madcat(
        data_spec: Data,
        feature_types: dict = None,
        text_handler: TextHandler = None, is_text_handler=True,
        time_series_handler: TimeSeriesHandler = None, is_time_series_handler=True,
        feature_stats: FeatureStats = None, is_feature_stats=True,
        text_feature_extraction: TextFeatureExtraction = None,
        is_pre_model=True,
        is_model_training=True, model_run: ModelRun = None,
        is_post_model=True,
        is_model_weaknesses=True,
        plot=True,
        max_stats_col_width=30,
):
    """
        Perform end-to-end machine learning modeling on the provided data.

        This function takes in a data set and performs a comprehensive machine learning modeling process.
        It first identifies the problem type (either regression or classification) and the feature types
        (numeric, categorical, boolean, or text) based on metadata provided.
        Then, it selects the appropriate machine learning models to run, performs cross-validation (CV),
        and returns the best model based on comparison of the results.
        In addition, the function also provides feature statistics, pre-model analysis, post-model analysis,
        and identifies any weaknesses in the model.
        """
    # Tell a story
    story = """Welcome to MadCat, the comprehensive machine learning and data analysis solution!
    \nThis module is designed to streamline the entire process of data analysis and machine learning modeling, 
    \nfrom start to finish.
    \nTo learn more about MadCat, visit https://github.com/amito-ds/chester.
    \nMadCat performs all necessary pre-processing steps to get your data ready for modeling.
    \nIt then trains and tests multiple models, selecting the best one based on results.
    \nThe results are visually displayed for easy interpretation.\n"""

    if is_feature_stats:
        story += "The feature stats will be calculated to understand the data better.\n"
    if is_pre_model:
        story += "The pre model process will be performed to prepare the data for the model.\n"
    if data_spec.target_column is not None:
        story += "A model will be trained and tested using the given data.\n"
    if is_post_model:
        story += "Post model analysis will be performed to evaluate the model performance.\n"
    if is_model_weaknesses:
        story += "The weaknesses of the model will be analyzed to understand how to improve it.\n"
    if plot:
        story += "The results will be plotted for visualization.\n"
    story += "The maximum column width for stats display is set to {}.\n".format(max_stats_col_width)
    print(story)

    # reset the file
    rc = ReportCollector(REPORT_PATH)
    with open(REPORT_PATH, 'w') as f:
        pass
    run_metadata_collector = {}

    # meta learn
    df = data_spec.df.sample(frac=1).reset_index(drop=True)
    data_info = DataInfo(data=df.sample(min(10000, len(df))), target=data_spec.target_column)
    data_info.calculate()
    data_info.data = df

    if feature_types is not None:
        data_info = FeatureTypes(feature_types).update_data_info(data_info=data_info)
    target_column = data_info.target

    print(chapter_title("meta learn"))
    print(data_info)
    run_metadata_collector["data info"] = data_info

    rc.save_text(text="Full report to analyze my data:\n")
    rc.save_object(obj=data_info, text="Data information:")
    ####################################################
    # Text handling
    # cleaning
    if is_text_handler:
        text_cleaner = None
        text_pre_process = None
        if text_handler is not None:
            text_cleaner = text_handler.text_cleaner
            text_pre_process = text_handler.text_pre_process

        cleaner = CleanerHandler(data_info=data_info, text_cleaner=text_cleaner)
        cleaner.transform()
        data_info = cleaner.data_info

        # keep a copy of all text after cleaning
        text_cols = data_info.feature_types_val["text"]
        clean_text_df = pd.DataFrame()
        if len(text_cols) > 0:
            pd.options.mode.chained_assignment = None
            clean_text_df = data_info.data[text_cols]
            clean_text_df.rename(columns={col: "clean_" + col for col in clean_text_df.columns}, inplace=True)

        # preprocessing
        pp = PreprocessHandler(data_info=data_info, text_pre_process=text_pre_process)
        pp.transform()
        data_info_original = data_info  # in case you need it
        data_info = pp.data_info

        # data_info_text_cleaning for later text analysis
        if len(text_cols) > 0:
            clean_text_df = pd.concat([df, clean_text_df], axis=1)
        run_metadata_collector["data info"] = data_info
    ####################################################

    ####################################################
    # Time series handling
    # is_time_series_handler
    ####################################################
    # Feat extract
    print(chapter_title('feature engineering'))
    rc.save_text(text="features engineering process for the data:")
    # Handle TS
    orig_di = data_info
    if is_time_series_handler:
        if time_series_handler is None:
            time_series_handler = TimeSeriesHandler()
        time_cols = data_info.feature_types_val["time"]
        if len(time_cols) > 0:
            ts_fe = TimeSeriesFeaturesExtraction(data_info=data_info, time_series_handler=time_series_handler)
            ts_fe.run()
            ts_fe.data_info.data = ts_fe.data_info.data. \
                                       loc[:, ~ts_fe.data_info.data.columns.duplicated()]  # drop col dups
            data_info = ts_fe.data_info
            time_series_handler = ts_fe.time_series_handler
            # orig_di = data_info

    # Continue the rest
    feat_hand = FeaturesHandler(data_info=data_info)
    if text_feature_extraction is not None:
        feat_hand.text_feature_extraction = text_feature_extraction
    feature_types, final_df = feat_hand.transform()
    if target_column is not None:
        final_df[target_column] = data_info.data[data_info.target]
    final_df = final_df.loc[:, ~final_df.columns.duplicated()]
    data_info.data = data_info.data.loc[:, ~data_info.data.columns.duplicated()]

    run_metadata_collector["data info"] = data_info
    run_metadata_collector["features data"] = final_df
    ####################################################
    # Feat Stats
    data_info_num_stats = None
    if is_feature_stats:
        rc.save_text(
            text="features statistics for the data. Analyzed by groups (text, numeric, categorical) if exists:")
        print(chapter_title('feature statistics'))
        if feature_stats is None:
            feature_stats = FeatureStats()
        sample_obs = feature_stats.sample_obs
        # sample data for stats
        final_df = final_df.loc[:, ~final_df.columns.duplicated()].sample(min(sample_obs, len(final_df)))
        data_info_num_stats = DataInfo(data=final_df, target=target_column)
        data_info_num_stats.calculate()

        num_cols = data_info_num_stats.feature_types_val["numeric"]
        cat_cols = data_info.feature_types_val["categorical"]
        text_cols = data_info.feature_types_val["text"]

        # plot trick
        plot_stats = True
        if plot:
            plot_stats = True
        else:
            if feature_stats is not None:
                if feature_stats.plot is not None:
                    plot_stats = feature_stats.plot
            else:
                plot_stats = True

        max_stats_col_width = max_stats_col_width or 30
        if len(num_cols) > 0:
            title = "\nNumerical Feature statistics:"
            print(title)
            rc.save_text(title)
            # NumericStats(data_info_num_stats, max_print=max_stats_col_width).run(plot=plot_stats)
        title = "\nCategorical Feature statistics:"
        print(title)
        rc.save_text(title)

        CategoricalStats(data_info, max_print=max_stats_col_width).run(plot=plot_stats)
        if len(text_cols) > 0:
            print("Text Feature Statistics")
            orig_df = data_info.data
            data_info.data = clean_text_df
            TextStats(data_info, text_spec=text_feature_extraction).run()
            data_info.data = orig_df

        all_features = list(set(chain.from_iterable(list(data_info.feature_types_val.values()))))
        ts_cols = [feat for feat in all_features if feat.startswith("ts_")]
        if len(ts_cols) > 0:
            print("Time Series Feature Statistics")
            TimeSeriesFeatureStatistics(data_info=data_info, ts_cols=ts_cols,
                                        time_series_handler=time_series_handler).run(plot=True)
    ####################################################
    # No target => no model! The story ends here
    if data_info.problem_type_val == "No target variable":
        return run_metadata_collector

    # Pre model
    if is_pre_model:
        final_df = final_df.loc[:, ~final_df.columns.duplicated()]
        rc.save_text("** model pre analysis report:")
        print(chapter_title('model pre analysis'))
        # label stats
        print("Label pre model analysis:")
        TargetPreModelAnalysis(data_info=data_info, time_series_handler=time_series_handler).run(plot)
        TimeSeriesPreModelAnalysis(data_info=data_info, time_series_handler=time_series_handler).run()
        # num, cat pre model
        if data_info_num_stats is None:
            data_info_num_stats = DataInfo(data=final_df, target=target_column)
            data_info_num_stats.calculate()
        NumericPreModelAnalysis(data_info_num_stats).run(plot)
        plt.close()
        # cat if found any
        cat_not_ts = data_info.feature_types_val["categorical"]
        if len(cat_not_ts) > 0:
            print("Categorical pre model analysis:")
            CategoricalPreModelAnalysis(data_info).run(plot)
    ####################################################
    # model
    if not is_model_training:
        return run_metadata_collector
    if is_model_training:
        print(chapter_title("model training"))
        rc.save_text("Training models and choosing the best one")
        # drop col dups
        data_info.feature_types_val["numeric"] = list(set(data_info.feature_types_val["numeric"]))
        data_info.feature_types_val["categorical"] = list(set(data_info.feature_types_val["categorical"]))
        # encode if needed
        if data_info.problem_type_val in ["Binary classification", "Multiclass classification"]:
            label_encoder = LabelEncoder()
            final_df[target_column] = label_encoder.fit_transform(final_df[target_column])

        # cv data
        cv_data = CVData(train_data=final_df, test_data=None, target_column='target', split_data=True)
        data_info.feature_types_val = feature_types
        # run the model
        if model_run is not None:
            model = BestModel(data_info=data_info,
                              cv_data=cv_data,
                              num_models_to_compare=model_run.n_models,
                              best_practice_prob=model_run.best_practice_prob)
        else:
            model = BestModel(data_info=data_info, cv_data=cv_data)
        model_results = model.get_best_model()  # returns result of the best baseline model
        # print model metadata
        params = model_results[1].get_params()
        try:
            print(f"Best model: {type(model_results[1].model)}, with parameters:")
            for p in params:
                print(p.name, ":", p.value)
        except:
            pass

        run_metadata_collector["data info"] = data_info
        run_metadata_collector["features data"] = final_df
        run_metadata_collector["model"] = model_results[1]
        run_metadata_collector["parameters"] = params
        run_metadata_collector["model_results"] = model_results[0]
    ####################################################

    # post model
    is_baseline = type(model_results[1]).__name__ == "BaselineModel"
    if is_baseline:
        print("Best model is a simple baseline")
        return run_metadata_collector
    if is_post_model and is_model_training:
        rc.save_text("\nPost model analysis - analyzing results of the chosen model: ")
        print(chapter_title('post model analysis'))
        post_model_analysis = PostModelAnalysis(cv_data, data_info, model=model_results[1])
        run_metadata_collector["post_model_analysis"] = post_model_analysis
        post_model_analysis.analyze()

        model_bootstrap = ModelBootstrap(cv_data, data_info, model=model_results[1])
        run_metadata_collector["model_bootstrap"] = model_bootstrap
        model_bootstrap.plot()
    ####################################################

    #  model weaknesses
    if is_model_weaknesses and is_model_training:
        rc.save_text("Trying to find weaknesses in the model by training models on the error:")
        print(chapter_title('model weaknesses'))
        model_weaknesses = ModelWeaknesses(cv_data, data_info, model=model_results[1])
        run_metadata_collector["model_weaknesses"] = model_weaknesses
        model_weaknesses.run()
    ####################################################

    return run_metadata_collector
